import os
from dataclasses import dataclass, field

import keras
import numpy as np
import tensorflow as tf
from packaging import version

try:
    import torch
except ImportError:
    torch = None

try:
    import torch_geometric
except ImportError:
    torch_geometric = None

# Check Keras version
if version.parse(keras.__version__).major >= 3:
    from keras import ops as kops
else:
    kops = None


@dataclass
class MLPSettings:
    """
    _summary_

    Args: (all optional, default values inside data class)
        embedding_layers (list): _description_
        numerical_dense_layers (list): _description_
        dense_layers (list): _description_
        dense_dropouts (list): _description_
    """

    # Default settings
    embedding_layers: list = field(default_factory=lambda: [16, 16, 16, 16])
    numerical_dense_layers: list = field(default_factory=lambda: [16])
    dense_layers: list = field(default_factory=lambda: [64, 32, 64, 32])
    dense_dropouts: list = field(
        default_factory=lambda: [
            0.0,
            0.2,
            0.2,
            0.0,
        ]
    )

    def to_config(self):
        return {
            "embedding_layers": self.embedding_layers,
            "numerical_dense_layers": self.numerical_dense_layers,
            "dense_layers": self.dense_layers,
            "dense_dropouts": self.dense_dropouts,
        }

    @classmethod
    def from_config(cls, config):
        return cls(**config)


@dataclass
class TransformerSettings:
    """
    _summary_

    Args: (all optional, default values inside data class)
        global_dense_layers (list): _description_
        seq_dense_layers (list): _description_

        global_numerical_dense_layers (list): _description_
        layer_numerical_dense_layers (list): _description_

        num_blocks (int): _description_
        num_heads (int): _description_
        ff_dim (int): _description_
        output_dim (int): _description_
        dropout_rate (int): _description_

        global_embedding_layers (list): _description_
        seq_embedding_layers (list): _description_
        dense_layers (list): _description_
        dense_dropouts (list): _description_
    """

    # Default settings
    global_dense_layers: list = field(default_factory=lambda: [64, 128])
    seq_dense_layers: list = field(default_factory=lambda: [64, 128])

    global_numerical_dense_layers: list = field(default_factory=lambda: [32])
    seq_numerical_dense_layers: list = field(default_factory=lambda: [32])

    num_blocks: int = 2
    num_heads: int = 4
    ff_dim: int = 128
    output_dim: int = 64
    dropout_rate: float = 0.1

    global_embedding_layers: list = field(default_factory=lambda: [16, 16, 16, 16])
    seq_embedding_layers: list = field(default_factory=lambda: [16, 16, 16, 16])

    dense_layers: list = field(default_factory=lambda: [128, 128, 64])
    dense_dropouts: list = field(
        default_factory=lambda: [
            0.2,
            0.2,
        ]
    )

    def to_config(self):
        return {
            "global_dense_layers": self.global_dense_layers,
            "seq_dense_layers": self.seq_dense_layers,
            "global_numerical_dense_layers": self.global_numerical_dense_layers,
            "seq_numerical_dense_layers": self.seq_numerical_dense_layers,
            "num_blocks": self.num_blocks,
            "num_heads": self.num_heads,
            "ff_dim": self.ff_dim,
            "output_dim": self.output_dim,
            "dropout_rate": self.dropout_rate,
            "global_embedding_layers": self.global_embedding_layers,
            "seq_embedding_layers": self.seq_embedding_layers,
            "dense_layers": self.dense_layers,
            "dense_dropouts": self.dense_dropouts,
        }

    @classmethod
    def from_config(cls, config):
        return cls(**config)


@dataclass
class GNNSettings:
    """
    _summary_

    Args: (all optional, default values inside data class)
    """

    # Default settings
    numerical_dense_layers: list = field(default_factory=lambda: [32])
    gconv_layers: list = field(default_factory=lambda: [128, 64])

    global_embedding_layers: list = field(default_factory=lambda: [16, 16, 16, 16])
    seq_embedding_layers: list = field(default_factory=lambda: [16, 16, 16, 16])

    dense_layers: list = field(default_factory=lambda: [128, 128, 64])
    bayesian_dense: list = field(default_factory=lambda: [False, False, False])
    dense_dropouts: list = field(
        default_factory=lambda: [
            0.2,
            0.2,
        ]
    )
    bayesian_output: bool = False

    def to_config(self):
        return {
            "numerical_dense_layers": self.numerical_dense_layers,
            "gconv_layers": self.gconv_layers,
            "global_embedding_layers": self.global_embedding_layers,
            "seq_embedding_layers": self.seq_embedding_layers,
            "dense_layers": self.dense_layers,
            "bayesian_dense": self.bayesian_dense,
            "dense_dropouts": self.dense_dropouts,
            "bayesian_output": self.bayesian_output,
        }

    @classmethod
    def from_config(cls, config):
        return cls(**config)


class KerasMLP(keras.Model):
    def __init__(
        self, settings: MLPSettings, input_shape, output_shape, categorical_maps, name="KerasMLP"
    ):
        squeeze_op = tf.squeeze if kops is None else kops.squeeze

        categorical_inputs = []
        embeddings = []

        for idx, key in enumerate(categorical_maps):
            # Keras 3 reorders the Input Tensors by name (tree.flatten), causing shape mismatches in fit() call
            # [g(lobal)|s(equential)]{idx} prefix helps keep the order deterministic
            input_layer = keras.layers.Input(shape=(1,), name=f"g{idx}_{key}_input")

            embedding_layer = squeeze_op(
                keras.layers.Embedding(
                    input_dim=len(categorical_maps[key]) + 1,
                    output_dim=settings.embedding_layers[idx],
                    name=f"{key}_embedding",
                )(input_layer),
                axis=-2,
            )

            categorical_inputs.append(input_layer)
            embeddings.append(embedding_layer)

        numerical_inputs = keras.layers.Input(
            shape=(input_shape[-1] - len(categorical_inputs),),
            name=f"g{len(categorical_maps)}_numerical_inputs",
        )

        numerical_outputs = numerical_inputs
        for idx, units in enumerate(settings.numerical_dense_layers):
            numerical_outputs = keras.layers.Dense(
                units, use_bias=False, name=f"numerical_dense_{idx}"
            )(numerical_outputs)

        concat_op = tf.concat if kops is None else kops.concatenate
        x = concat_op([*embeddings, numerical_outputs], axis=-1)

        for idx, units in enumerate(settings.dense_layers):
            x = keras.layers.Dense(units, activation="relu")(x)
            if idx < len(settings.dense_dropouts) and settings.dense_dropouts[idx] > 0.0:
                x = KerasMCDropout(settings.dense_dropouts[idx])(x)

        model_inputs = categorical_inputs + [numerical_inputs]
        model_outputs = keras.layers.Dense(output_shape[-1], activation="relu")(x)

        super().__init__(inputs=model_inputs, outputs=model_outputs, name=name)
        super().build([None] + list(input_shape))

        self.settings = settings

        self.global_input_keys = {
            "categorical": [layer.name for layer in categorical_inputs],
            "numerical": numerical_inputs.name,
        }
        self.global_input_shape = input_shape
        self.global_categorical_maps = categorical_maps

    def to_config(self):
        return {
            "name": self.name,
            "settings": self.settings.to_config(),
            "input_shape": self.global_input_shape,
            "output_shape": self.output_shape,
            "categorical_maps": self.global_categorical_maps,
        }

    @classmethod
    def from_config(cls, config):
        config["settings"] = MLPSettings.from_config(config["settings"])
        return cls(**config)


class KerasTransformer(keras.Model):
    def __init__(
        self,
        settings: TransformerSettings,
        global_input_shape,
        sequential_input_shape,
        output_shape,
        global_categorical_maps,
        sequential_categorical_maps,
        name="KerasTransformer",
    ):
        squeeze_op = tf.squeeze if kops is None else kops.squeeze
        concat_op = tf.concat if kops is None else kops.concatenate

        # Global inputs (global model/hls info)
        global_categorical_inputs = []
        global_embeddings = []

        for idx, key in enumerate(global_categorical_maps):
            input_layer = keras.layers.Input(shape=(1,), name=f"g{idx}_{key}_input")

            embedding_layer = squeeze_op(
                keras.layers.Embedding(
                    input_dim=len(global_categorical_maps[key]) + 1,
                    output_dim=settings.global_embedding_layers[idx],
                    name=f"{key}_embedding",
                )(input_layer),
                axis=-2,
            )

            global_categorical_inputs.append(input_layer)
            global_embeddings.append(embedding_layer)

        global_numerical_inputs = keras.layers.Input(
            shape=(global_input_shape[-1] - len(global_categorical_maps),),
            name=f"g{len(global_categorical_maps)}_numerical_inputs",
        )

        global_numerical_outputs = global_numerical_inputs
        for idx, units in enumerate(settings.global_numerical_dense_layers):
            global_numerical_outputs = keras.layers.Dense(units, use_bias=False)(
                global_numerical_outputs
            )

        concat_global_inputs = concat_op([*global_embeddings, global_numerical_outputs], axis=-1)

        global_outputs = concat_global_inputs
        for idx, units in enumerate(settings.global_dense_layers):
            global_outputs = keras.layers.Dense(units, activation="relu")(global_outputs)

        # Sequential inputs (layer-wise info)
        sequential_categorical_inputs = []
        sequential_embeddings = []

        for idx, key in enumerate(sequential_categorical_maps):
            input_layer = keras.layers.Input(
                shape=(
                    None,
                    1,
                ),
                name=f"s{idx}_{key}_input",
            )

            embedding_layer = squeeze_op(
                ZeroMaskEmbedding(
                    input_dim=len(sequential_categorical_maps[key]) + 1,
                    output_dim=settings.seq_embedding_layers[idx],
                    name=f"{key}_embedding",
                )(input_layer),
                axis=-2,
            )

            sequential_categorical_inputs.append(input_layer)
            sequential_embeddings.append(embedding_layer)

        sequential_numerical_inputs = keras.layers.Input(
            shape=(None, sequential_input_shape[-1] - len(sequential_categorical_inputs)),
            name=f"s{len(sequential_categorical_maps)}_numerical_inputs",
        )

        sequential_numerical_outputs = sequential_numerical_inputs
        for idx, units in enumerate(settings.seq_numerical_dense_layers):
            sequential_numerical_outputs = keras.layers.Dense(units, use_bias=False)(
                sequential_numerical_outputs
            )

        concat_seq_inputs = concat_op(
            [*sequential_embeddings, sequential_numerical_outputs], axis=-1
        )
        masked_inputs = keras.layers.Masking(mask_value=0.0)(concat_seq_inputs)
        mask = tf.reduce_any(tf.not_equal(masked_inputs, 0.0), axis=-1)  # shape: (batch, seq_len)

        x = masked_inputs
        for idx, units in enumerate(settings.seq_dense_layers):
            x = keras.layers.TimeDistributed(keras.layers.Dense(units, activation="relu"))(x)

        x = KerasTransformerBlock(
            x.shape[-1],
            settings.num_heads,
            settings.ff_dim,
            settings.output_dim,
            settings.dropout_rate,
        )(x, mask=mask)

        # seq_lens = tf.reduce_sum(tf.cast(mask, tf.int32), axis=1)
        # indices = tf.stack([tf.range(tf.shape(x)[0]), seq_lens - 1], axis=1)
        # x = tf.gather_nd(x, indices)

        x = keras.layers.Concatenate(axis=-1)([x, global_outputs])

        for idx, units in enumerate(settings.dense_layers):
            x = keras.layers.Dense(units, activation="relu")(x)
            if idx < len(settings.dense_dropouts) and settings.dense_dropouts[idx] > 0.0:
                x = KerasMCDropout(settings.dense_dropouts[idx])(x)

        # The same order as the model inputs should be maintained, else Keras 3 breaks
        inputs = (
            global_categorical_inputs
            + [global_numerical_inputs]
            + sequential_categorical_inputs
            + [sequential_numerical_inputs]
        )
        outputs = keras.layers.Dense(output_shape[-1], activation="relu")(x)

        super().__init__(inputs=inputs, outputs=outputs, name=name)
        super().build([None] + list(global_input_shape) + list(sequential_input_shape))

        self.settings = settings

        self.global_input_keys = {
            "categorical": [layer.name for layer in global_categorical_inputs],
            "numerical": global_numerical_inputs.name,
        }
        self.global_input_shape = global_input_shape
        self.global_categorical_maps = global_categorical_maps

        self.sequential_input_keys = {
            "categorical": [layer.name for layer in sequential_categorical_inputs],
            "numerical": sequential_numerical_inputs.name,
        }
        self.sequential_input_shape = sequential_input_shape
        self.sequential_categorical_maps = sequential_categorical_maps

    def to_config(self):
        return {
            "name": self.name,
            "settings": self.settings.to_config(),
            "global_input_shape": self.global_input_shape,
            "sequential_input_shape": self.sequential_input_shape,
            "output_shape": self.output_shape,
            "global_categorical_maps": self.global_categorical_maps,
            "sequential_categorical_maps": self.sequential_categorical_maps,
        }

    @classmethod
    def from_config(cls, config):
        config["settings"] = TransformerSettings.from_config(config["settings"])
        return cls(**config)


class TorchMLP(torch.nn.Module):
    def __init__(
        self,
        settings: MLPSettings,
        input_shape,
        output_shape,
        categorical_maps,
        name="TorchMLP",
        device=None,
    ):
        super().__init__()

        self.embeddings = torch.nn.ModuleDict()
        for idx, key in enumerate(categorical_maps):
            self.embeddings[f"g{idx}_{key}_embedding"] = torch.nn.Embedding(
                num_embeddings=len(categorical_maps[key]) + 1,
                embedding_dim=settings.embedding_layers[idx],
            )

        self.numerical_layers = torch.nn.ModuleDict()
        in_dim = input_shape[-1] - len(categorical_maps)
        if in_dim > 0:
            for idx, units in enumerate(settings.numerical_dense_layers):
                self.numerical_layers[f"g{idx + len(categorical_maps)}_numerical_layer"] = (
                    torch.nn.Linear(in_dim, units, bias=False)
                )
                in_dim = units

        concat_dim = sum(settings.embedding_layers) + in_dim
        self.dense_layers = torch.nn.ModuleList()
        self.dropouts = torch.nn.ModuleList()
        for idx, units in enumerate(settings.dense_layers):
            self.dense_layers.append(torch.nn.Linear(concat_dim, units))
            if idx < len(settings.dense_dropouts) and settings.dense_dropouts[idx] > 0.0:
                self.dropouts.append(torch.nn.Dropout(settings.dense_dropouts[idx]))
            else:
                self.dropouts.append(None)

            concat_dim = units

        self.output_layer = torch.nn.Linear(concat_dim, output_shape[-1])

        self.name = name
        self.settings = settings

        self.global_input_keys = {
            "categorical": [name for name in self.embeddings.keys()],
            "numerical": f"g{len(categorical_maps)}_numerical_layer",
        }
        self.global_input_shape = input_shape
        self.global_categorical_maps = categorical_maps

        self.output_shape = output_shape

        if device is None:
            if torch.cuda.is_available():
                device = torch.device("cuda")
            else:
                device = torch.device("cpu")
        self.device = device
        self.to(self.device)

    def forward(self, inputs):
        x_categorical = []
        for idx, key in enumerate(self.embeddings.keys()):
            x_categorical.append(self.embeddings[key](inputs[key]).squeeze(1))

        x_numerical = inputs[self.global_input_keys["numerical"]]
        for idx, key in enumerate(self.numerical_layers.keys()):
            x_numerical = self.numerical_layers[key](x_numerical)

        x = torch.cat(x_categorical + [x_numerical], dim=-1)
        for dense, dropout in zip(self.dense_layers, self.dropouts):
            x = torch.nn.functional.relu(dense(x))
            if dropout:
                x = dropout(x)

        outputs = self.output_layer(x)
        outputs = torch.nn.functional.softplus(outputs)
        # outputs = torch.nn.functional.relu(outputs)
        return outputs

    def to_config(self):
        return {
            "name": self.name,
            "settings": self.settings.to_config(),
            "input_shape": self.global_input_shape,
            "output_shape": self.output_shape,
            "categorical_maps": self.global_categorical_maps,
        }

    @classmethod
    def from_config(cls, config):
        config["settings"] = MLPSettings.from_config(config["settings"])
        return cls(**config)


class TorchGNN(torch.nn.Module):
    def __init__(
        self,
        settings: GNNSettings,
        global_input_shape,
        sequential_input_shape,
        output_shape,
        global_categorical_maps,
        sequential_categorical_maps,
        name="TorchGNN",
        device=None,
    ):
        if torch_geometric is None:
            raise ImportError(
                "Failed to import \"torch_geometric\". Please install \"torch_geometric\" to use this class."
            )

        super().__init__()

        self.global_embeddings = torch.nn.ModuleDict()
        for idx, key in enumerate(global_categorical_maps):
            self.global_embeddings[f"g{idx}_{key}_embedding"] = torch.nn.Embedding(
                num_embeddings=len(global_categorical_maps[key]) + 1,
                embedding_dim=settings.global_embedding_layers[idx],
            )

        self.sequential_embeddings = torch.nn.ModuleDict()
        for idx, key in enumerate(sequential_categorical_maps):
            self.sequential_embeddings[f"s{idx}_{key}_embedding"] = torch.nn.Embedding(
                num_embeddings=len(sequential_categorical_maps[key]) + 1,
                embedding_dim=settings.seq_embedding_layers[idx],
            )

        self.numerical_layers = torch.nn.ModuleDict()
        g_in_dim = max(0, global_input_shape[-1] - len(global_categorical_maps))
        if g_in_dim > 0:
            for idx, units in enumerate(settings.numerical_dense_layers):
                self.numerical_layers[f"g{idx + len(global_categorical_maps)}_numerical_layer"] = (
                    torch.nn.Linear(g_in_dim, units, bias=False)
                )
                g_in_dim = units

        seq_numerical_dim = sequential_input_shape[-1] - len(sequential_categorical_maps)
        gnn_dim = sum(settings.seq_embedding_layers) + seq_numerical_dim

        # Project global features to the same dimension as the GNN node features?
        g_token_dim = sum(settings.global_embedding_layers) + g_in_dim
        self.g_projection = torch.nn.Sequential(
            torch.nn.Linear(g_token_dim, gnn_dim),
            torch.nn.ReLU(),
        )

        self.gconvs = torch.nn.ModuleList()
        self.norms = torch.nn.ModuleList()
        for idx, units in enumerate(settings.gconv_layers):
            mlp = torch.nn.Sequential(
                torch.nn.Linear(gnn_dim, units), torch.nn.ReLU(), torch.nn.Linear(units, units)
            )
            self.gconvs.append(
                # torch_geometric.nn.SAGEConv(
                #     in_channels=gnn_dim,
                #     out_channels=units,
                #     bias=True,
                # )
                torch_geometric.nn.GINConv(nn=mlp)
            )
            self.norms.append(torch_geometric.nn.GraphNorm(in_channels=units))
            gnn_dim = units
        self.jumping_knowledge = torch_geometric.nn.JumpingKnowledge(mode="cat")

        self.attention_pool = torch_geometric.nn.AttentionalAggregation(
            gate_nn=torch.nn.Sequential(torch.nn.Linear(sum(settings.gconv_layers), 1))
        )
        self.global_pool = torch_geometric.nn.global_mean_pool

        concat_dim = sum(settings.global_embedding_layers) + g_in_dim + sum(settings.gconv_layers)
        self.dense_layers = torch.nn.ModuleList()
        self.dropouts = torch.nn.ModuleList()
        for idx, units in enumerate(settings.dense_layers):
            dense_layer = (
                TorchBayesianLinear(concat_dim, units)
                if (idx < len(settings.bayesian_dense) and settings.bayesian_dense[idx])
                else torch.nn.Linear(concat_dim, units)
            )
            self.dense_layers.append(dense_layer)
            if idx < len(settings.dense_dropouts) and settings.dense_dropouts[idx] > 0.0:
                self.dropouts.append(torch.nn.Dropout(settings.dense_dropouts[idx]))
            else:
                self.dropouts.append(None)

            concat_dim = units

        self.output_layer = (
            TorchBayesianLinear(concat_dim, output_shape[-1])
            if settings.bayesian_output
            else torch.nn.Linear(concat_dim, output_shape[-1])
        )

        self.name = name
        self.settings = settings

        self.global_input_keys = {
            "categorical": [name for name in self.global_embeddings.keys()],
            "numerical": f"g{len(global_categorical_maps)}_numerical_layer",
        }
        self.global_input_shape = global_input_shape
        self.global_categorical_maps = global_categorical_maps

        self.sequential_input_keys = {
            "categorical": [name for name in self.sequential_embeddings.keys()],
            "numerical": f"s{len(sequential_categorical_maps)}_numerical_layer",
        }
        self.sequential_input_shape = sequential_input_shape
        self.sequential_categorical_maps = sequential_categorical_maps

        self.output_shape = output_shape
        self.kl_loss = 0.0

        if device is None:
            if torch.cuda.is_available() and os.environ.get("CUDA_VISIBLE_DEVICES") not in [
                "",
                "-1",
            ]:
                device = torch.device("cuda")
            else:
                device = torch.device("cpu")
        self.device = device
        self.to(self.device)

    def forward(self, inputs):
        x_global_categorical = []
        for idx, key in enumerate(self.global_embeddings.keys()):
            try:
                x_global_categorical.append(self.global_embeddings[key](inputs[key]).squeeze(1))
            except Exception as e:
                print(f"Input: {inputs[key].shape}")
                raise e

        x_global_numerical = inputs[self.global_input_keys["numerical"]]
        for idx, key in enumerate(self.numerical_layers.keys()):
            x_global_numerical = self.numerical_layers[key](x_global_numerical)

        seq_numerical_inputs = inputs[
            self.sequential_input_keys["numerical"]
        ]  # shape: (batch, seq_len, seq_features)
        mask = ~(seq_numerical_inputs == 0.0).all(dim=-1)  # mask to eliminate zero padding

        x_sequential_categorical = []
        for idx, key in enumerate(self.sequential_embeddings.keys()):
            x_sequential_categorical.append(self.sequential_embeddings[key](inputs[key]).squeeze(2))

        node_features = []
        edge_indices = []
        batch_vector = []
        ptr = 0
        for i in range(seq_numerical_inputs.shape[0]):
            valid_seq_numerical = seq_numerical_inputs[i][mask[i]]
            if valid_seq_numerical.size(0) == 0:
                continue

            valid_seq_embeddings = [embedding[i][mask[i]] for embedding in x_sequential_categorical]
            valid_features = torch.cat([*valid_seq_embeddings, valid_seq_numerical], dim=-1)
            node_features.append(valid_features.to(self.device))
            num_nodes = valid_features.size(0)
            if num_nodes > 1:
                row = torch.arange(num_nodes - 1, dtype=torch.long)
                edge = torch.stack([row, row + 1], dim=0) + ptr
                edge_indices.append(edge.to(self.device))

            batch_vector.append(torch.full((num_nodes,), i, dtype=torch.long).to(self.device))
            ptr += num_nodes

        if len(node_features) == 0:
            raise ValueError("All sequential samples are empty after masking.")

        x_seq = torch.cat(node_features, dim=0)
        edge_indices = (
            torch.cat(edge_indices, dim=1)
            if edge_indices
            else torch.empty((2, 0), dtype=torch.long)
        )
        batch = torch.cat(batch_vector, dim=0)

        if edge_indices.numel() > 0:
            edge_indices = torch_geometric.utils.to_undirected(edge_indices, num_nodes=ptr)
        edge_indices, _ = torch_geometric.utils.add_self_loops(edge_indices, num_nodes=ptr)

        xs = []
        for gconv, norm in zip(self.gconvs, self.norms):
            x_seq = gconv(x_seq, edge_index=edge_indices)
            x_seq = norm(x_seq)
            x_seq = torch.nn.functional.relu(x_seq)
            xs.append(x_seq)
        x_seq = self.jumping_knowledge(xs)

        pooled_seq = self.attention_pool(x_seq, index=batch)
        # pooled_seq = self.global_pool(x_seq, batch=batch)

        self.kl_loss = 0.0
        x = torch.cat([*x_global_categorical, x_global_numerical, pooled_seq], dim=-1)
        for idx, dense in enumerate(self.dense_layers):
            x = torch.nn.functional.relu(dense(x))
            if hasattr(dense, "kl_divergence"):
                self.kl_loss += dense.kl_divergence()
            if self.dropouts[idx] is not None:
                x = self.dropouts[idx](x)

        outputs = self.output_layer(x)
        if hasattr(self.output_layer, "kl_divergence"):
            self.kl_loss += self.output_layer.kl_divergence()
        outputs = torch.nn.functional.softplus(outputs)
        # outputs = torch.nn.functional.relu(outputs)
        return outputs

    def to_config(self):
        return {
            "name": self.name,
            "settings": self.settings.to_config(),
            "global_input_shape": self.global_input_shape,
            "sequential_input_shape": self.sequential_input_shape,
            "output_shape": self.output_shape,
            "global_categorical_maps": self.global_categorical_maps,
            "sequential_categorical_maps": self.sequential_categorical_maps,
        }

    @classmethod
    def from_config(cls, config):
        config["settings"] = GNNSettings.from_config(config["settings"])
        return cls(**config)


class KerasTransformerBlock(keras.layers.Layer):
    """
    _summary_

    Args:
        embed_dim (_type_): _description_
        num_heads (_type_): _description_
        ff_dim (_type_): _description_
        output_dim (_type_): _description_
        dropout_rate (_type_): _description_. Defaults to 0.1
    """

    def __init__(
        self, embed_dim, num_heads, ff_dim, output_dim, dropout_rate=0.1, global_pool="last"
    ):
        super().__init__()

        if global_pool not in ["avg", "last"]:
            raise ValueError(
                f"global_pool must be one of [\"avg\", \"last\"], but got \"{global_pool}\""
            )
        self.global_pool = global_pool

        self.att = keras.layers.MultiHeadAttention(num_heads=num_heads, key_dim=embed_dim)
        self.ffn = keras.Sequential(
            [
                keras.layers.Dense(ff_dim, activation="relu"),
                keras.layers.Dense(embed_dim),
            ]
        )
        self.layernorm1 = keras.layers.LayerNormalization(epsilon=1e-6)
        self.layernorm2 = keras.layers.LayerNormalization(epsilon=1e-6)
        self.dropout1 = KerasMCDropout(dropout_rate)
        self.dropout2 = KerasMCDropout(dropout_rate)

        self.global_avg_pooling = keras.layers.GlobalAveragePooling1D()
        self.out_dense = keras.layers.Dense(output_dim)

        self.supports_masking = True  # needed to pass the mask to downstream layers

    def call(self, inputs, training=False, mask=None):
        attention_mask = None
        if mask is not None:
            attention_mask = tf.cast(tf.expand_dims(mask, axis=1), dtype=tf.float32)

        seq_len = tf.shape(inputs)[1]
        pos_encoding = self.positional_encoding(seq_len, inputs.shape[-1])
        inputs = inputs + pos_encoding

        attn_output = self.att(inputs, inputs, attention_mask=attention_mask, training=training)
        attn_output = self.dropout1(attn_output, training=training)
        out1 = self.layernorm1(inputs + attn_output)
        ffn_output = self.ffn(out1)
        ffn_output = self.dropout2(ffn_output, training=training)
        outputs = self.layernorm2(out1 + ffn_output)

        if self.global_pool == "avg":
            mask = tf.cast(mask, outputs.dtype)
            mask = tf.expand_dims(mask, axis=-1)
            outputs = tf.reduce_sum(outputs * mask, axis=1) / (tf.reduce_sum(mask, axis=1) + 1e-6)
        elif self.global_pool == "last":
            seq_lens = tf.reduce_sum(tf.cast(mask, tf.int32), axis=1)
            indices = tf.stack([tf.range(tf.shape(outputs)[0]), seq_lens - 1], axis=1)
            outputs = tf.gather_nd(outputs, indices)

        outputs = self.out_dense(outputs)
        return outputs

    def compute_mask(self, inputs, mask=None):
        return mask

    # Positional encoding doesn't seem to be important
    def positional_encoding(self, position, d_model):
        angle_rads = self._get_angles(
            tf.range(position, dtype=tf.float32)[:, tf.newaxis],
            tf.range(d_model, dtype=tf.float32)[tf.newaxis, :],
            d_model,
        )

        sines = tf.sin(angle_rads[:, 0::2])
        cosines = tf.cos(angle_rads[:, 1::2])

        pos_encoding = tf.concat([sines, cosines], axis=-1)
        return tf.expand_dims(pos_encoding, axis=0)

    def _get_angles(self, pos, i, d_model):
        angle_rates = 1 / tf.pow(10000.0, (2 * (i // 2)) / tf.cast(d_model, tf.float32))
        return pos * angle_rates


class ZeroMaskEmbedding(keras.layers.Layer):
    def __init__(self, input_dim, output_dim, **kwargs):
        """
        Custom embedding layer that zeroes out and freezes the 0th index embedding vector.
        Args:
            input_dim (int): Size of the vocabulary.
            output_dim (int): Dimension of the embedding vectors.
        """

        super().__init__(**kwargs)
        self.embedding = keras.layers.Embedding(
            input_dim=input_dim,
            output_dim=output_dim,
            mask_zero=False,  # We'll manually handle masking
        )
        self.output_dim = output_dim

    def build(self, input_shape):
        super().build(input_shape)

        # Manually zero out the 0th embedding vector
        self.embedding.build(input_shape)
        embeddings = self.embedding.embeddings
        embeddings.assign(
            tf.tensor_scatter_nd_update(
                embeddings, indices=[[0]], updates=[tf.zeros(self.output_dim)]
            )
        )

    def call(self, inputs):
        return self.embedding(inputs)


class KerasMCDropout(keras.layers.Dropout):
    def __init__(self, rate, **kwargs):
        """
        Expose an additional attribute to better control the call method's
        'training' argument for MC Dropout at inference time.

        Args:
            rate (_type_): _description_
        """
        self.enable_mc = kwargs.pop("enable_mc", False)
        super().__init__(rate, **kwargs)

    def call(self, inputs, training=None):
        # Normal behavior during training:
        #   - when training=True (from model.fit), keep it.
        #   - when training=False, dropout would normally be off (assuming enable_mc=False).
        # During inference, if enable_mc is True, we force training=True to enable MC dropout.

        if not training and self.enable_mc:
            training = True
        return super().call(inputs, training=training)


class TorchBayesianLinear(torch.nn.Module):
    def __init__(self, in_features, out_features, prior_std=1.0):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features

        self.weight_mu = torch.nn.Parameter(torch.Tensor(out_features, in_features))
        self.weight_logvar = torch.nn.Parameter(torch.Tensor(out_features, in_features))
        self.bias_mu = torch.nn.Parameter(torch.Tensor(out_features))
        self.bias_logvar = torch.nn.Parameter(torch.Tensor(out_features))

        # Prior N(0, prior_std^2)
        self.prior_std = prior_std
        self.prior_logvar = 2 * torch.log(torch.tensor(prior_std))
        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.kaiming_uniform_(self.weight_mu, a=np.sqrt(5))
        torch.nn.init.constant_(self.weight_logvar, -5)  # small initial variance

        fan_in, _ = torch.nn.init._calculate_fan_in_and_fan_out(self.weight_mu)
        bound = 1 / np.sqrt(fan_in)
        torch.nn.init.uniform_(self.bias_mu, -bound, bound)
        torch.nn.init.constant_(self.bias_logvar, -5)

    def forward(self, input):
        # Sample weights using reparameterization trick
        weight_std = torch.exp(0.5 * self.weight_logvar)
        bias_std = torch.exp(0.5 * self.bias_logvar)
        weight_eps = torch.randn_like(self.weight_mu)
        bias_eps = torch.randn_like(self.bias_mu)
        weight = self.weight_mu + weight_std * weight_eps
        bias = self.bias_mu + bias_std * bias_eps

        return torch.nn.functional.linear(input, weight, bias)

    def kl_divergence(self):
        """KL(q(w|θ) || p(w)) for both weights and biases, assuming Gaussian prior."""
        # posterior and prior variances
        w_var = torch.exp(self.weight_logvar)
        b_var = torch.exp(self.bias_logvar)
        prior_var = torch.exp(self.prior_logvar)

        # KL for each Gaussian param: 0.5 * [ (σ² + μ²)/prior_var - 1 - log(σ²/prior_var) ]
        kl_weight = (
            0.5
            * (
                (w_var + self.weight_mu**2) / prior_var
                - (self.weight_logvar - self.prior_logvar)
                - 1.0
            ).sum()
        )
        kl_bias = (
            0.5
            * (
                (b_var + self.bias_mu**2) / prior_var - (self.bias_logvar - self.prior_logvar) - 1.0
            ).sum()
        )

        return kl_weight + kl_bias



import math
import os

try:
    import torch
except ImportError:
    torch = None


class ReduceLROnPlateau:
    def __init__(
        self, monitor="val_loss", mode="min", factor=0.5, patience=5, min_delta=0.0, min_lr=1e-6
    ):
        self.monitor = monitor
        self.mode = mode
        self.factor = factor
        self.patience = patience
        self.min_delta = min_delta
        self.min_lr = min_lr

        self._best = math.inf if mode == "min" else -math.inf
        self._current_lrs = []
        self._num_bad = 0

    def _has_improved(self, value) -> bool:
        if self.mode == "min":
            return value < (self._best - self.min_delta)
        else:
            return value > (self._best + self.min_delta)

    def step(self, value, wrapper, verbose=0):
        if not self._current_lrs:
            self._current_lrs = [
                param_group['lr'] for param_group in wrapper.optimizer.param_groups
            ]

        if self._has_improved(value):
            self._best = float(value)
            self._num_bad = 0
            return

        self._num_bad += 1
        if self._num_bad >= self.patience:
            for idx in range(len(wrapper.optimizer.param_groups)):
                old_lr = wrapper.optimizer.param_groups[idx]['lr']
                new_lr = max(old_lr * self.factor, self.min_lr)
                if old_lr > self.min_lr:
                    wrapper.optimizer.param_groups[idx]['lr'] = new_lr
                    self._current_lrs[idx] = new_lr
                    if verbose > 0:
                        print(
                            f"[{self.__class__.__name__}] Reducing learning rate "
                            + f"from {old_lr:.6f} to {new_lr:.6f}."
                        )
            self._num_bad = 0


class EarlyStopping:
    def __init__(
        self, monitor="val_loss", mode="min", patience=10, min_delta=0.0, restore_best=True
    ):
        self.monitor = monitor
        self.mode = mode
        self.patience = patience
        self.min_delta = min_delta
        self.restore_best = restore_best

        self._best = math.inf if mode == "min" else -math.inf
        self._num_bad = 0
        self._best_state = None

    def _has_improved(self, value) -> bool:
        if self.mode == "min":
            return value < (self._best - self.min_delta)
        else:
            return value > (self._best + self.min_delta)

    def step(self, value, wrapper, verbose=0) -> bool:
        """Returns True if training should stop."""
        if self._has_improved(value):
            self._best = float(value)
            self._num_bad = 0
            if self.restore_best:
                self._best_state = {
                    k: v.detach().cpu().clone() for k, v in wrapper.model.state_dict().items()
                }
            return False

        self._num_bad += 1
        if self._num_bad >= self.patience:
            if verbose > 0:
                print(
                    f"[{self.__class__.__name__}] Early stopping triggered. "
                    + f"Best {self.monitor}: {self._best:.6f}"
                )
            if self.restore_best and self._best_state is not None:
                wrapper.model.load_state_dict(self._best_state)
                if verbose > 0:
                    print(f"[{self.__class__.__name__}] Restored best model state.")
            return True
        return False


class ModelCheckpoint:
    def __init__(self, dirpath, monitor="val_loss", mode="min"):
        if torch is None:
            raise ImportError(
                "torch is not installed. Please install torch to use ModelCheckpoint."
            )
        assert os.path.isdir(dirpath), f"{dirpath} is not a valid directory"
        assert mode in ["min", "max"], "mode should be 'min' or 'max'"

        self.dirpath = dirpath
        self.monitor = monitor
        self.mode = mode

        self._best = math.inf if mode == "min" else -math.inf

    def _has_improved(self, value) -> bool:
        if self.mode == "min":
            return value < self._best
        else:
            return value > self._best

    def step(self, value, wrapper, verbose=0):
        if self._has_improved(value):
            self._best = float(value)
            payload = {
                "best_value": self._best,
                "monitor": self.monitor,
                "model": wrapper.model.state_dict(),
                "optimizer": wrapper.optimizer.state_dict(),
            }

            os.makedirs(self.dirpath, exist_ok=True)
            filepath = os.path.join(self.dirpath, f"{wrapper.model.name}_best.pth")
            torch.save(payload, filepath)
            wrapper.save(self.dirpath)  # save weights and config separately too

            if verbose > 0:
                print(
                    f"[{self.__class__.__name__}] {self.monitor} improved to {self._best:.6f}. "
                    + f"Saved to {self.dirpath}"
                )



import itertools
import json
import os
import time
from dataclasses import dataclass, field

import keras
import numpy as np
import pandas as pd
import tensorflow as tf
from sklearn.metrics import r2_score
from tqdm.auto import tqdm

try:
    import torch
except ImportError:
    torch = None

try:
    import onnx
except ImportError:
    onnx = None

from rule4ml.models.architectures import (
    KerasMCDropout,
    KerasMLP,
    KerasTransformer,
    TorchGNN,
    TorchMLP,
)
from rule4ml.models.callbacks import EarlyStopping
from rule4ml.models.metrics import rmse, smape
from rule4ml.models.scaling import *  # noqa: F403
from rule4ml.models.utils import get_loss_from_str, get_optimizer_from_str
from rule4ml.parsers.data_parser import (
    boards_data,
    get_global_inputs,
    get_layers_data,
    to_dataframe,
)
from rule4ml.parsers.network_parser import (
    config_from_keras_model,
    config_from_onnx_model,
    config_from_torch_model,
)
from rule4ml.parsers.utils import camel_keys_to_snake, fixed_precision_to_bit_width, to_lower_keys


@dataclass
class TrainSettings:
    """
    _summary_

    Args: (all optional, default values inside data class)
        num_epochs (int): _description_
        batch_size (int): _description_
        learning_rate (float): _description_

        optimizer (str): _description_
        loss_function (str): _description_
        metrics (list): _description_
    """

    # Default settings

    num_epochs: int = 50
    batch_size: int = 32
    learning_rate: float = 1e-4

    optimizer: str = "adam"
    loss_function: str = "mae"
    metrics: list = field(default_factory=lambda: ["mape"])


class BaseModelWrapper:
    def __init__(self):
        self.reset()

    def reset(self):
        self.global_input_keys = {"categorical": [], "numerical": ""}
        self.sequential_input_keys = {"categorical": [], "numerical": ""}

        self.global_numerical_labels = []
        self.sequential_numerical_labels = []

        self.global_categorical_maps = {}
        self.sequential_categorical_maps = {}

        self.output_labels = []
        self.output_shape = ()

        if hasattr(self, "model") and self.model is not None:
            del self.model

        self.model = None
        self.dataset = None
        self.scaler = None

    def set_model(self, model):
        self.model = model

        self.global_input_keys = getattr(
            model, "global_input_keys", {"categorical": [], "numerical": ""}
        )
        self.sequential_input_keys = getattr(
            model, "sequential_input_keys", {"categorical": [], "numerical": ""}
        )

        self.global_categorical_maps = getattr(model, "global_categorical_maps", {})
        self.sequential_categorical_maps = getattr(model, "sequential_categorical_maps", {})

        self.output_shape = getattr(model, "output_shape", ())

    def build_inputs(self, inputs_df):
        if self.model is None:
            raise Exception("A model needs to be set or loaded before building inputs.")

        input_dict = {}
        for key, label in zip(
            self.global_input_keys["categorical"], self.global_categorical_maps.keys()
        ):
            input_dict[key] = inputs_df[label].values

        global_numerical_inputs_df = inputs_df.drop(
            self.global_categorical_maps.keys(), axis=1
        ).select_dtypes(
            exclude=[object]
        )  # excluding nested "sequential_inputs" Series
        self.global_numerical_labels = global_numerical_inputs_df.columns.values

        input_dict[self.global_input_keys["numerical"]] = global_numerical_inputs_df.values

        if "sequential_inputs" in inputs_df:
            seq_inputs_series = inputs_df["sequential_inputs"]
            if not seq_inputs_series.empty:
                for key, label in zip(
                    self.sequential_input_keys["categorical"],
                    self.sequential_categorical_maps.keys(),
                ):
                    input_dict[key] = seq_inputs_series.apply(lambda df: df[label]).values

                seq_numerical_inputs = seq_inputs_series.apply(
                    lambda df: df.drop(self.sequential_categorical_maps.keys(), axis=1)
                    .select_dtypes(exclude=[object])
                    .values
                ).values
                seq_numerical_inputs = np.asarray([arr for arr in seq_numerical_inputs])

                self.sequential_numerical_labels = (
                    seq_inputs_series.iloc[0]
                    .drop(self.sequential_categorical_maps.keys(), axis=1)
                    .select_dtypes(exclude=[object])
                    .columns.values
                )

                input_dict[self.sequential_input_keys["numerical"]] = seq_numerical_inputs

        return input_dict

    def build_dataset(
        self,
        inputs_df,
        targets_df,
        batch_size,
        val_inputs_df=None,
        val_targets_df=None,
        scaler=None,
    ):
        if self.model is None:
            raise Exception("A model needs to be set or loaded before building a dataset.")

        self.batch_size = batch_size
        self.output_labels = targets_df.columns.values

        if scaler:
            targets_df = pd.DataFrame(
                [
                    scaler.transform(target_row, input_row)
                    for target_row, input_row in zip(
                        targets_df.to_dict(orient="records"), inputs_df.to_dict(orient="records")
                    )
                ],
                index=targets_df.index,
            )
            if val_targets_df is not None:
                val_targets_df = pd.DataFrame(
                    [
                        scaler.transform(target_row, input_row)
                        for target_row, input_row in zip(
                            val_targets_df.to_dict(orient="records"),
                            val_inputs_df.to_dict(orient="records"),
                        )
                    ],
                    index=val_targets_df.index,
                )
        self.scaler = scaler
        return targets_df, val_targets_df

    def fit(self):
        if self.model is None:
            raise Exception("A model needs to be set or loaded before training.")
        if self.dataset is None:
            raise Exception("A dataset needs to be built before training a model.")
        return

    def predict_from_df(self, inputs_df, targets=None, mc_dropout=False, verbose=0):
        if self.model is None:
            raise Exception("A model needs to be set or loaded before making predictions.")

        feature_labels = list(self.global_numerical_labels) + list(
            self.global_categorical_maps.keys()
        )
        if len(self.sequential_numerical_labels) > 0:
            sequential_labels = list(self.sequential_numerical_labels) + list(
                self.sequential_categorical_maps.keys()
            )
            inputs_df.loc[:, "sequential_inputs"] = inputs_df["sequential_inputs"].apply(
                lambda x: x[sequential_labels]
            )
            feature_labels += ["sequential_inputs"]

        inputs_df = inputs_df[feature_labels]
        inputs = self.build_inputs(inputs_df)

        start_time = time.time()
        prediction = self._model_predict(inputs, 0, mc_dropout=mc_dropout, verbose=verbose)
        total_time = time.time() - start_time

        if self.scaler:
            prediction = np.asarray(
                [
                    list(
                        self.scaler.inverse(
                            dict(zip(self.output_labels, pred_row)),
                            dict(zip(inputs_df.columns, inputs_df.iloc[idx])),
                        ).values()
                    )
                    for idx, pred_row in enumerate(prediction)
                ]
            )

        if verbose > 1:
            r2 = r2_score(targets, prediction, force_finite=False)
            print(f"R2 Score: {r2:.2f}")

            smape_value = smape(targets, prediction)
            print(f"SMAPE: {smape_value:.2f}%")

            rmse_value = rmse(targets, prediction)
            print(f"RMSE: {rmse_value:.2f}")

            avg_inference_time = total_time / len(inputs_df)
            print(f"Average Inference Time: {avg_inference_time:.2E} seconds")

            return prediction, r2, smape_value, rmse_value, avg_inference_time

        return prediction

    def predict(
        self,
        models_to_predict,
        hls_configs,
        hls4ml_versions,
        vivado_versions,
        mc_dropout=False,
        verbose=0,
    ):
        if self.model is None:
            raise Exception("A model needs to be set or loaded before making predictions.")

        if not isinstance(models_to_predict, (tuple, list)):
            raise TypeError("models_to_predict expects a list of keras or torch models.")
        if not isinstance(hls_configs, (tuple, list)):
            raise TypeError("hls_configs expects a list of HLS configuration dictionaries.")
        if not isinstance(hls4ml_versions, (tuple, list)):
            raise TypeError("hls4ml_versions expects a list or tuple of version strings.")
        if not isinstance(vivado_versions, (tuple, list)):
            raise TypeError("vivado_versions expects a list or tuple of version strings.")

        supported_categoricals = {}
        for key, mapping in self.global_categorical_maps.items():
            supported_categoricals[key] = list(to_lower_keys(mapping).keys())

        accepted_hls_configs = []
        for hls_config in hls_configs:
            board = hls_config.get("board", "").lower()
            strategy = hls_config.get("model", {}).get("strategy", "").lower()
            if board not in supported_categoricals.get("board", []):
                raise Exception(f"Model {self.model.name} does not support board: {board}")
            if strategy not in supported_categoricals.get("strategy", []):
                raise Exception(f"Model {self.model.name} does not support strategy: {strategy}")
            accepted_hls_configs.append(hls_config)
        if not accepted_hls_configs:
            accepted_hls_configs = [None]

        accepted_hls4ml_versions = []
        for hls4ml_version in hls4ml_versions:
            if hls4ml_version:
                if hls4ml_version.lower() not in supported_categoricals.get("hls4ml_version", []):
                    raise Exception(
                        f"Model {self.model.name} does not support hls4ml_version: {hls4ml_version}"
                    )
                accepted_hls4ml_versions.append(hls4ml_version)
        if not accepted_hls4ml_versions:
            accepted_hls4ml_versions = [None]

        accepted_vivado_versions = []
        for vivado_version in vivado_versions:
            if vivado_version:
                if vivado_version.lower() not in supported_categoricals.get("vivado_version", []):
                    raise Exception(
                        f"Model {self.model.name} does not support vivado_version: {vivado_version}"
                    )
                accepted_vivado_versions.append(vivado_version)
        if not accepted_vivado_versions:
            accepted_vivado_versions = [None]

        global_inputs = []
        sequential_inputs = []

        for model_to_predict, hls_config, hls4ml_version, vivado_version in itertools.product(
            models_to_predict,
            accepted_hls_configs,
            accepted_hls4ml_versions,
            accepted_vivado_versions,
        ):
            if isinstance(model_to_predict, keras.Model):
                model_config = config_from_keras_model(model_to_predict, hls_config)
            elif torch is not None and isinstance(model_to_predict, torch.nn.Module):
                model_config = config_from_torch_model(model_to_predict, hls_config)
            elif onnx is not None and isinstance(model_to_predict, onnx.ModelProto):
                model_config = config_from_onnx_model(model_to_predict, hls_config)
            else:
                if torch is not None:
                    if onnx is not None:
                        raise TypeError(f"Unsupported model type: {type(model_to_predict)}")
                    else:
                        raise ImportError("Failed to import module onnx.")
                elif onnx is not None:
                    raise ImportError("Failed to import module torch.")

            global_inputs.append(
                get_global_inputs(
                    model_config,
                    hls_config,
                    hls4ml_version=hls4ml_version,
                    vivado_version=vivado_version,
                )
            )
            sequential_inputs.append(get_layers_data(model_config))

        inputs_df = to_dataframe(
            global_inputs=global_inputs,
            sequential_inputs=sequential_inputs,
            global_categorical_maps=self.global_categorical_maps,
            sequential_categorical_maps=self.sequential_categorical_maps,
            meta_data=[],
            targets=[],
        )

        return self.predict_from_df(inputs_df, mc_dropout=mc_dropout, verbose=verbose)

    def mc_predict_from_df(self, inputs_df, num_samples=50, return_all=False, verbose=0):
        """
        Perform MC Dropout predictions from a dataframe.

        Args:
            inputs_df (pd.DataFrame): input dataframe.
            num_samples (int, optional): number of samples for MC Dropout. Defaults to 100.
            return_all (bool, optional): whether to return all samples or just the mean and std. Defaults to False.
            verbose (int, optional): verbosity level. Defaults to 0.

        Returns:
            tuple: mean and std of predictions if return_all is False, else mean, std, and all samples.
        """
        if self.model is None:
            raise Exception("A model needs to be set or loaded before making predictions.")

        all_predictions = []
        for _ in range(num_samples):
            predictions = self.predict_from_df(inputs_df, mc_dropout=True, verbose=verbose)
            all_predictions.append(predictions)

        all_predictions = np.array(all_predictions)
        predictions_mean = np.mean(all_predictions, axis=0)
        predictions_std = np.std(all_predictions, axis=0)

        outputs = (predictions_mean, predictions_std)
        if return_all:
            outputs += (all_predictions,)

        return outputs

    def mc_predict(
        self, models_to_predict, hls_configs, num_samples=50, return_all=False, verbose=0
    ):
        """
        Perform MC Dropout predictions.

        Args:
            models_to_predict (list): list of models to predict.
            hls_configs (list): list of HLS configuration dictionaries.
            num_samples (int, optional): number of samples for MC Dropout. Defaults to 100.
            return_all (bool, optional): whether to return all samples or just the mean and std. Defaults to False.
            verbose (int, optional): verbosity level. Defaults to 0.

        Returns:
            tuple: mean and std of predictions if return_all is False, else mean, std, and all samples.
        """
        if self.model is None:
            raise Exception("A model needs to be set or loaded before making predictions.")

        all_predictions = []
        for _ in range(num_samples):
            predictions = self.predict(
                models_to_predict, hls_configs, mc_dropout=True, verbose=verbose
            )
            all_predictions.append(predictions)

        all_predictions = np.array(all_predictions)
        predictions_mean = np.mean(all_predictions, axis=0)
        predictions_std = np.std(all_predictions, axis=0)

        outputs = (predictions_mean, predictions_std)
        if return_all:
            outputs += (all_predictions,)

        return outputs

    def to_config(self):
        return {
            "global_numerical_labels": list(self.global_numerical_labels),
            "sequential_numerical_labels": list(self.sequential_numerical_labels),
            "global_categorical_maps": self.global_categorical_maps,
            "sequential_categorical_maps": self.sequential_categorical_maps,
            "output_labels": list(self.output_labels),
            "output_shape": self.output_shape,
            "model_class": self.model.__class__.__name__,
            "model_config": self.model.to_config(),
            "scaler_class": self.scaler.__class__.__name__ if self.scaler else None,
            "scaler_config": self.scaler.to_config() if self.scaler else None,
        }

    def from_config(self, config):
        self.global_numerical_labels = config["global_numerical_labels"]
        self.sequential_numerical_labels = config["sequential_numerical_labels"]
        self.global_categorical_maps = config["global_categorical_maps"]
        self.sequential_categorical_maps = config["sequential_categorical_maps"]

        self.output_labels = config["output_labels"]
        self.output_shape = config["output_shape"]

        model_class = config["model_class"]
        self.model_config = config["model_config"]
        self.model = globals()[model_class].from_config(self.model_config)
        self.set_model(self.model)

        scaler_class = config.get("scaler_class", None)
        scaler_config = config.get("scaler_config", None)
        if scaler_class and scaler_config:
            self.scaler = globals()[scaler_class].from_config(scaler_config)

    def save(self, save_dir):
        """
        Save the class config and model weights.

        Args:
            save_dir (_type_): _description_
        """

        if not os.path.exists(save_dir):
            os.makedirs(save_dir)

        config = self.to_config()
        config_path = os.path.join(save_dir, f"{self.model.name}.config.json")
        with open(config_path, "w") as config_file:
            json.dump(config, config_file, indent=2)

        self._save_weights(save_dir)

    def load(self, config_path, weights_path):
        """
        Load saved class config.

        Args:
            config_path (_type_): _description_
        """

        if not config_path.endswith(".json"):
            raise ValueError("Config file name must end with .json")
        if not os.path.isfile(config_path):
            raise FileNotFoundError(f"Config file does not exist: {config_path}")
        if not os.path.isfile(weights_path):
            raise FileNotFoundError(f"Weights file does not exist: {weights_path}")

        self.reset()

        with open(config_path) as json_file:
            # Load the class config
            config = json.load(json_file)
            self.from_config(config)

        self._load_weights(weights_path)

    def _enable_mc_dropout(self):
        raise NotImplementedError("This method should be implemented in subclasses.")

    def _model_predict(self, inputs, batch_size, mc_dropout, verbose=0):
        raise NotImplementedError("This method should be implemented in subclasses.")

    def _save_weights(self, save_dir):
        raise NotImplementedError("This method should be implemented in subclasses.")

    def _load_weights(self, weights_path):
        raise NotImplementedError("This method should be implemented in subclasses.")


class TorchModelWrapper(BaseModelWrapper):
    def __init__(self):
        super().__init__()
        if torch is None:
            raise ImportError("Torch import failed. Please install torch to use this wrapper.")

    def set_model(self, model):
        super().set_model(model)
        self.device = model.device

    def _process_input_dict(self, input_dict):
        input_tensors = []
        for key, value in input_dict.items():
            if (
                key in self.global_input_keys["categorical"]
                or key in self.sequential_input_keys["categorical"]
            ):
                # Convert categorical inputs to long tensors
                input_tensors.append(torch.tensor(value, dtype=torch.long))
            else:
                # Convert numerical inputs to float tensors
                input_tensors.append(torch.tensor(value, dtype=torch.float32))

        return input_tensors

    def build_dataset(
        self,
        inputs_df,
        targets_df,
        batch_size,
        val_ratio=0.2,
        val_inputs_df=None,
        val_targets_df=None,
        train_repeats=1,
        shuffle=True,
        scaler=None,
        verbose=0,
    ):
        targets_df, val_targets_df = super().build_dataset(
            targets_df, batch_size, val_targets_df=val_targets_df, scaler=scaler
        )

        input_dict = self.build_inputs(inputs_df)
        if verbose > 0:
            for key, value in input_dict.items():
                print(f"{key}: {value.shape}")

        input_tensors = self._process_input_dict(input_dict)
        target_tensors = torch.tensor(targets_df.values)

        self.dataset = torch.utils.data.TensorDataset(*input_tensors, target_tensors)
        if val_inputs_df is not None and val_targets_df is not None:
            indices = torch.randperm(len(self.dataset))
            train_indices = indices.tolist() * train_repeats
            train_subset = torch.utils.data.Subset(self.dataset, train_indices)

            val_input_dict = self.build_inputs(val_inputs_df)
            if verbose > 0:
                for key, value in val_input_dict.items():
                    print(f"val_{key}: {value.shape}")

            val_target_tensors = torch.tensor(val_targets_df.values)
            val_input_tensors = []
            for key, value in val_input_dict.items():
                if (
                    key in self.global_input_keys["categorical"]
                    or key in self.sequential_input_keys["categorical"]
                ):
                    val_input_tensors.append(torch.tensor(value, dtype=torch.long))
                else:
                    val_input_tensors.append(torch.tensor(value, dtype=torch.float32))
            val_subset = torch.utils.data.TensorDataset(*val_input_tensors, val_target_tensors)
        else:
            total_size = len(inputs_df)
            val_size = int(total_size * val_ratio)

            indices = torch.randperm(total_size)
            val_indices = indices[:val_size].tolist()
            train_indices = indices[val_size:].tolist() * train_repeats
            train_subset = torch.utils.data.Subset(self.dataset, train_indices)
            val_subset = torch.utils.data.Subset(self.dataset, val_indices)

        def collate_fn(batch):
            batch = list(zip(*batch))
            inputs = {key: torch.stack(batch[i]) for i, key in enumerate(input_dict.keys())}
            targets = torch.stack(batch[-1]).to(dtype=torch.float32)
            return inputs, targets

        self.train_data = torch.utils.data.DataLoader(
            train_subset,
            batch_size=batch_size,
            shuffle=shuffle,
            collate_fn=collate_fn,
        )
        self.val_data = torch.utils.data.DataLoader(
            val_subset,
            batch_size=batch_size,
            shuffle=False,
            collate_fn=collate_fn,
        )

    def fit(self, train_settings: TrainSettings, callbacks=[], verbose=0):
        super().fit()

        self.optimizer = train_settings.optimizer
        if isinstance(self.optimizer, str):
            self.optimizer = get_optimizer_from_str(
                self.optimizer,
                backend="torch",
                params=self.model.parameters(),
                lr=train_settings.learning_rate,
            )

        loss_function = train_settings.loss_function
        if isinstance(loss_function, str):
            loss_function = get_loss_from_str(loss_function, backend="torch")

        metric_functions = train_settings.metrics

        history = {
            "train": {
                "loss": [],
                **{m.name: [] for m in metric_functions},
            },
        }
        if self.val_data:
            history["val"] = {
                "loss": [],
                **{m.name: [] for m in metric_functions},
            }

        num_batches = len(self.train_data)
        for epoch in range(train_settings.num_epochs):
            self.model.train()
            train_loss = 0.0

            with tqdm(
                self.train_data,
                desc=f"Epoch {epoch + 1}/{train_settings.num_epochs}",
                leave=False,
            ) as pbar:
                for inputs, targets in pbar:
                    inputs = {key: value.to(self.device) for key, value in inputs.items()}
                    targets = targets.to(self.device)

                    self.optimizer.zero_grad()
                    outputs = self.model(inputs)

                    loss = loss_function(outputs, targets)
                    if hasattr(self.model, "kl_loss"):
                        loss += 1 / num_batches * self.model.kl_loss

                    loss.backward()
                    self.optimizer.step()

                    train_loss += loss.item()
                    for m in metric_functions:
                        m.update(targets.detach(), outputs.detach(), self.output_labels)

                    postfix = {"loss": f"{(train_loss / (pbar.n + 1)):.6f}"}
                    for m in metric_functions:
                        current_metric = m.result().item()
                        postfix[m.name] = f"{current_metric:.6f}"

                    pbar.set_postfix(postfix)

            avg_train_loss = train_loss / len(self.train_data)
            history["train"]["loss"].append(avg_train_loss)
            for m in metric_functions:
                history["train"][m.name].append(m.result().item())
                m.reset()  # Reset metrics for validation or next epoch

            summary = f"Epoch {epoch + 1}/{train_settings.num_epochs}\n"
            summary += f"loss: {avg_train_loss:.6f} -"
            for m in metric_functions:
                summary += f" {m.name}: {history['train'][m.name][-1]:.6f} -"

            if self.val_data:
                self.model.eval()
                val_loss = 0.0

                with torch.no_grad():
                    for inputs, targets in self.val_data:
                        inputs = {key: value.to(self.device) for key, value in inputs.items()}
                        targets = targets.to(self.device)

                        outputs = self.model(inputs)

                        loss = loss_function(outputs, targets)
                        val_loss += loss.item()
                        for m in metric_functions:
                            m.update(targets, outputs, self.output_labels)

                avg_val_loss = val_loss / len(self.val_data)
                history["val"]["loss"].append(avg_val_loss)
                for m in metric_functions:
                    history["val"][m.name].append(m.result().item())
                    m.reset()

                summary += f" val_loss: {avg_val_loss:.6f} -"
                for m in metric_functions:
                    summary += f" val_{m.name}: {history['val'][m.name][-1]:.6f} -"

            if verbose > 0:
                print(summary)

            for callback in callbacks:
                if not hasattr(callback, "step"):
                    raise NotImplementedError(
                        f"Callback {callback} does not implement a step() method."
                    )
                monitor = callback.monitor
                monitor_parts = monitor.split("_")
                monitor_value = None
                if len(monitor_parts) > 1 and monitor_parts[0] in history:
                    monitor = "_".join(monitor_parts[1:])
                    monitor_value = history[monitor_parts[0]].get(monitor, None)
                if monitor_value is None:
                    raise Exception(f"Monitored value {callback.monitor} not found in fit history.")

                cb_result = callback.step(monitor_value[-1], self, verbose=verbose)
                if isinstance(callback, EarlyStopping) and cb_result:
                    return history

        return history

    def _enable_mc_dropout(self):
        """
        Enable MC Dropout by setting all dropout layers to train mode.
        """
        if self.model is None:
            raise Exception("A model needs to be set or loaded before enabling MC Dropout.")

        def set_dropout_train(m):
            if isinstance(m, torch.nn.Dropout):
                m.train()

        self.model.apply(set_dropout_train)

    def _model_predict(self, inputs, batch_size, mc_dropout, verbose=0):
        self.model.eval()
        if mc_dropout:
            self._enable_mc_dropout()

        with torch.no_grad():
            if isinstance(inputs, dict):
                input_tensors = self._process_input_dict(inputs)
                inputs = {
                    key: value.to(self.device) for key, value in zip(inputs.keys(), input_tensors)
                }

            outputs = self.model(inputs)
            return outputs.detach().cpu().numpy()

    def _save_weights(self, save_dir):
        weights_path = os.path.join(save_dir, f"{self.model.name}_weights.pt")
        torch.save(self.model.state_dict(), weights_path)

    def _load_weights(self, weights_path):
        if not weights_path.endswith(".pt"):
            raise ValueError("Weights filename must end with .pt")

        self.model.load_state_dict(torch.load(weights_path, map_location=self.device))


class KerasModelWrapper(BaseModelWrapper):
    def __init__(self):
        super().__init__()

    def build_dataset(
        self,
        inputs_df,
        targets_df,
        batch_size,
        val_ratio=0.2,
        val_inputs_df=None,
        val_targets_df=None,
        train_repeats=1,
        shuffle=True,
        scaler=None,
        verbose=0,
    ):
        targets_df, val_targets_df = super().build_dataset(
            targets_df, batch_size, val_targets_df=val_targets_df, scaler=scaler
        )

        input_dict = self.build_inputs(inputs_df)
        if verbose > 0:
            for key, value in input_dict.items():
                print(f"{key}: shape={value.shape}, dtype={value.dtype}")

        self.dataset = tf.data.Dataset.from_tensor_slices((input_dict, np.array(targets_df.values)))
        if val_inputs_df is not None and val_targets_df is not None:
            val_input_dict = self.build_inputs(val_inputs_df)

            train_data = self.dataset.repeat(train_repeats)
            if shuffle:
                train_data = train_data.shuffle(len(train_data))
            train_data = train_data.batch(batch_size)

            val_data = tf.data.Dataset.from_tensor_slices(
                (val_input_dict, np.array(val_targets_df.values))
            )
            val_data = val_data.batch(batch_size)

            self.train_data = train_data
            self.val_data = val_data
        elif val_ratio > 0.0:
            val_len = int(len(self.dataset) * val_ratio)
            val_data = self.dataset.take(val_len)
            val_data = val_data.batch(batch_size)
            train_data = self.dataset.skip(val_len).repeat(train_repeats)
            if shuffle:
                train_data = train_data.shuffle(len(self.dataset) - val_len)
            train_data = train_data.batch(batch_size)

            self.train_data = train_data
            self.val_data = val_data
        else:
            train_data = self.dataset.repeat(train_repeats)
            if shuffle:
                train_data = train_data.shuffle(len(train_data))
            train_data = train_data.batch(batch_size)

            self.train_data = train_data
            self.val_data = None

    def fit(self, train_settings: TrainSettings, callbacks=[], verbose=0):
        super().fit()

        self.optimizer = train_settings.optimizer
        if isinstance(self.optimizer, str):
            self.optimizer = get_optimizer_from_str(
                self.optimizer, backend="keras", learning_rate=train_settings.learning_rate
            )

        loss_function = train_settings.loss_function
        if isinstance(loss_function, str):
            loss_function = get_loss_from_str(loss_function, backend="keras")

        self.model.compile(
            optimizer=self.optimizer,
            loss=loss_function,
            metrics=train_settings.metrics,
        )

        return self.model.fit(
            self.train_data,
            epochs=train_settings.num_epochs,
            batch_size=train_settings.batch_size,
            validation_data=self.val_data,
            callbacks=callbacks,
            verbose=verbose,
        )

    def _enable_mc_dropout(self):
        """
        Enable MC Dropout by setting all dropout layers to train mode.
        """
        if self.model is None:
            raise Exception("A model needs to be set or loaded before enabling MC Dropout.")

        for layer in self.model.layers:
            if isinstance(layer, KerasMCDropout):
                layer.training = True

    def _model_predict(self, inputs, batch_size, mc_dropout, verbose=0):
        if mc_dropout:
            self._enable_mc_dropout()
        return self.model.predict(inputs, batch_size, verbose=verbose)

    def _save_weights(self, save_dir):
        weights_path = os.path.join(save_dir, f"{self.model.name}.weights.h5")
        self.model.save_weights(weights_path)

    def _load_weights(self, weights_path):
        if not weights_path.endswith(".weights.h5"):
            raise ValueError("Weights filename must end with .weights.h5")

        self.model.load_weights(weights_path)


class MultiModelWrapper:
    """
    _summary_
    """

    _default_precisions = ["ap_fixed<2, 1>", "ap_fixed<8, 3>", "ap_fixed<16, 6>"]
    _default_reuse_factors = [1, 2, 4, 8, 16, 32, 64]

    def __init__(self):
        self._models: dict[str, BaseModelWrapper] = {}

    def add_model_wrapper(self, model_wrapper: BaseModelWrapper):
        key = "-".join([x.upper() for x in model_wrapper.output_labels])
        self._models[key] = model_wrapper

    def load_default_models(self):
        base_path = os.path.join(os.path.dirname(__file__), "weights", "v2", "gnn")
        default_paths = [
            os.path.join(base_path, "BRAM"),
            os.path.join(base_path, "DSP"),
            os.path.join(base_path, "FF"),
            os.path.join(base_path, "LUT"),
            os.path.join(base_path, "CYCLES"),
            os.path.join(base_path, "INTERVAL"),
        ]
        for path in default_paths:
            if not os.path.exists(path + ".config.json"):
                continue

            with open(f"{path}.config.json") as f:
                config = json.load(f)
            class_name = config.get("model_class", None)
            if class_name in [KerasMLP.__name__, KerasTransformer.__name__]:
                if os.path.exists(f"{path}.weights.h5"):
                    model_wrapper = KerasModelWrapper()
                    model_wrapper.load(f"{path}.config.json", f"{path}.weights.h5")
                    self.add_model_wrapper(model_wrapper)
                else:
                    raise FileNotFoundError(f"Weights file not found: {path}.weights.h5")
            elif class_name in [TorchMLP.__name__, TorchGNN.__name__]:
                if os.path.exists(f"{path}.weights.pt"):
                    model_wrapper = TorchModelWrapper()
                    model_wrapper.load(f"{path}.config.json", f"{path}.weights.pt")
                    self.add_model_wrapper(model_wrapper)
                else:
                    raise FileNotFoundError(f"Weights file not found: {path}.weights.pt")
            else:
                raise NotImplementedError(f"Model class not supported: {class_name}")

    def predict(self, models_to_predict, **kwargs):
        """
        Predict FPGA implementation metrics for one or more models across one or more
        configuration(s), and return the results as a tabular `pandas.DataFrame`.

        This method evaluates every combination of:
        - each entry in `models_to_predict`,
        - each entry in `hls_configs` (either provided or auto-generated from defaults or values seen during training),
        - each entry in `hls4ml_versions` (either provided or auto-generated from defaults or values seen during training),
        - each entry in `vivado_versions` (either provided or auto-generated from defaults or values seen during training)

        Parameters
        ----------
        models_to_predict : A Keras/Torch model or a list/tuple of such models
            A single model or a collection of models to evaluate.

        **kwargs
            Optional keyword arguments:

            hls_configs : dict or list[dict], optional
                One configuration dictionary or a list of configuration dictionaries describing the
                target board and hls4ml model settings. Keys may be in camelCase or snake_case; they
                are normalized internally.

                Expected structure (minimum):
                {
                    "board": "<board_name>",
                    "model": {
                        "strategy": "<strategy>",
                        "precision": "<precision>",
                        "reuse_factor": <int>
                    }
                }

                If omitted, configurations are generated from defaults and values seen during each estimator's training.

            hls4ml_versions : str or list[str] or tuple[str], optional
                The hls4ml version(s) to associate with the model(s) being predicted.

            vivado_versions : str or list[str] or tuple[str], optional
                The Vivado version(s) to associate with the model(s) being predicted.

            round_digits : int, default=2
                Number of decimal places to round the output DataFrame values to.

            verbose : int, default=0
                Verbosity level forwarded to each estimator's `predict` call.

        Returns
        -------
        pandas.DataFrame
            A DataFrame of predictions with one row per (model, hls_config) pair.

        Raises
        ------
        TypeError
            If `hls_configs` is provided but is not a dict or a list/tuple of dicts.

        ValueError
            If a configuration specifies a board that is not present in `boards_data`.
        """

        if not self._models:
            raise Exception("No models have been added to the MultiModelWrapper.")

        if not isinstance(models_to_predict, (tuple, list)):
            models_to_predict = [models_to_predict]

        hls_configs = kwargs.get("hls_configs", None)
        hls_configs_by_estimator = {}
        if hls_configs is None:
            for estimator_key, estimator_model in self._models.items():
                estimator_boards = self._get_categorical_values(estimator_model, "board")
                estimator_strategies = self._get_categorical_values(estimator_model, "strategy")
                hls_configs_by_estimator[estimator_key] = [
                    {
                        "model": {
                            "strategy": strategy,
                            "precision": precision,
                            "reuse_factor": reuse_factor,
                        },
                        "board": board,
                    }
                    for board, strategy, precision, reuse_factor in itertools.product(
                        estimator_boards,
                        estimator_strategies,
                        MultiModelWrapper._default_precisions,
                        MultiModelWrapper._default_reuse_factors,
                    )
                ]
        else:
            if not isinstance(hls_configs, (tuple, list)):
                if not isinstance(hls_configs, dict):
                    raise TypeError(
                        "hls_configs argument expects a dictionary or list of dictionaries. See docstring example for correct formatting."
                    )
                hls_configs = [hls_configs]
            hls_configs_by_estimator = {key: list(hls_configs) for key in self._models.keys()}

        for estimator_configs in hls_configs_by_estimator.values():
            for idx in range(len(estimator_configs)):
                estimator_configs[idx] = camel_keys_to_snake(estimator_configs[idx])

        hls4ml_versions = kwargs.get("hls4ml_versions", None)
        hls4ml_versions_by_estimator = {}
        if hls4ml_versions is None:
            for estimator_key, estimator_model in self._models.items():
                hls4ml_versions_by_estimator[estimator_key] = self._get_categorical_values(
                    estimator_model, "hls4ml_version"
                )
        else:
            if isinstance(hls4ml_versions, str):
                hls4ml_versions = [hls4ml_versions]
            hls4ml_versions_by_estimator = {
                key: list(hls4ml_versions) for key in self._models.keys()
            }

        vivado_versions = kwargs.get("vivado_versions", None)
        vivado_versions_by_estimator = {}
        if vivado_versions is None:
            for estimator_key, estimator_model in self._models.items():
                vivado_versions_by_estimator[estimator_key] = self._get_categorical_values(
                    estimator_model, "vivado_version"
                )
        else:
            if isinstance(vivado_versions, str):
                vivado_versions = [vivado_versions]
            vivado_versions_by_estimator = {
                key: list(vivado_versions) for key in self._models.keys()
            }

        outputs = []
        outputs_by_key: dict[str, dict] = {}
        for key, estimator_model in self._models.items():
            estimator_hls_configs = hls_configs_by_estimator[key] or [None]
            estimator_hls4ml_versions = hls4ml_versions_by_estimator[key] or [None]
            estimator_vivado_versions = vivado_versions_by_estimator[key] or [None]
            estimator_product = list(
                itertools.product(
                    models_to_predict,
                    estimator_hls_configs,
                    estimator_hls4ml_versions,
                    estimator_vivado_versions,
                )
            )

            for model_to_predict, hls_config, hls4ml_version, vivado_version in estimator_product:
                model_name = getattr(model_to_predict, "name", model_to_predict.__class__.__name__)
                output_key = (
                    model_name,
                    hls_config["board"],
                    hls_config["model"]["strategy"],
                    hls_config["model"]["precision"],
                    hls_config["model"]["reuse_factor"],
                    hls4ml_version,
                    vivado_version,
                )
                if output_key not in outputs_by_key:
                    outputs_by_key[output_key] = {
                        "Model": model_name,
                        "Board": hls_config["board"],
                        "Strategy": hls_config["model"]["strategy"],
                        "Precision": hls_config["model"]["precision"],
                        "Reuse Factor": hls_config["model"]["reuse_factor"],
                        "HLS4ML Version": hls4ml_version,
                        "Vivado Version": vivado_version,
                    }
                    outputs.append(outputs_by_key[output_key])

            estimator_predictions = estimator_model.predict(
                models_to_predict,
                estimator_hls_configs,
                estimator_hls4ml_versions,
                estimator_vivado_versions,
                verbose=kwargs.get("verbose", 0),
            ).astype(float)

            target_labels = key.split("-")
            for idx, model_predictions in enumerate(estimator_predictions):
                model_to_predict, hls_config, hls4ml_version, vivado_version = estimator_product[
                    idx
                ]
                model_name = getattr(model_to_predict, "name", model_to_predict.__class__.__name__)
                output_key = (
                    model_name,
                    hls_config["board"],
                    hls_config["model"]["strategy"],
                    hls_config["model"]["precision"],
                    hls_config["model"]["reuse_factor"],
                    hls4ml_version,
                    vivado_version,
                )
                output_row = outputs_by_key[output_key]
                predictions = {}
                for k in range(len(model_predictions)):
                    predictions[target_labels[k]] = model_predictions[k]
                    available_key = f"max_{target_labels[k].lower()}"
                    if available_key in boards_data[output_row["Board"]]:
                        predictions[f"{target_labels[k]} (%)"] = (
                            model_predictions[k] / boards_data[output_row["Board"]][available_key]
                        ) * 100
                output_row.update(predictions)

        outputs_df = pd.DataFrame(outputs).round(kwargs.get("round_digits", 2))
        if not outputs_df.empty:
            sorted_boards = sorted(
                outputs_df["Board"].unique(), key=lambda x: list(boards_data.keys()).index(x)
            )
            outputs_df["Board"] = pd.Categorical(
                outputs_df["Board"], categories=sorted_boards, ordered=True
            )

            sorted_precisions = sorted(
                outputs_df["Precision"].unique(), key=lambda x: fixed_precision_to_bit_width(x)[0]
            )
            outputs_df["Precision"] = pd.Categorical(
                outputs_df["Precision"], categories=sorted_precisions, ordered=True
            )

        return outputs_df

    def _get_categorical_values(self, model_wrapper, key, fallback=[]):
        categorical_maps = getattr(model_wrapper, "global_categorical_maps", {}) or {}
        if key in categorical_maps:
            map_keys = list(categorical_maps[key].keys())
            return [k.lower() if isinstance(k, str) else k for k in map_keys]

        return list(fallback)



import math

import numpy as np

try:
    import torch
except Exception:
    torch = None


def _is_torch(x):
    return (torch is not None) and isinstance(x, torch.Tensor)


class BaseScaler:
    def transform(self, target_row, input_row):
        raise NotImplementedError()

    def inverse(self, target_row, input_row):
        raise NotImplementedError()

    def to_config(self):
        raise NotImplementedError()

    @classmethod
    def from_config(cls, config):
        raise NotImplementedError()


class BoardScaler(BaseScaler):
    def __init__(self, available_resources: dict, board_map: dict[str, int]):
        self.available_resources = {}
        self.board_map = board_map
        for board, resources in available_resources.items():
            board = self._get_board_name(board)
            self.available_resources[board] = {}
            for name, value in resources.items():
                self.available_resources[board][f"{name.lower()}"] = float(value)

    def _get_board_name(self, board):
        if isinstance(board, (int, float)):
            board = int(board)
            for name, idx in self.board_map.items():
                if idx == board:
                    return name.lower()
        elif isinstance(board, str):
            return board.lower()
        return board

    def transform(self, target_row, input_row):
        out = {}
        board_name = input_row.get("board")
        board_name = self._get_board_name(board_name)
        if board_name in self.available_resources:
            available = self.available_resources[board_name]
            for k, v in target_row.items():
                out[k] = v / available.get(k.lower(), 1.0)
        return out

    def inverse(self, target_row, input_row):
        out = {}
        board_name = input_row.get("board")
        board_name = self._get_board_name(board_name)
        if board_name in self.available_resources:
            available = self.available_resources[board_name]
            for k, z in target_row.items():
                out[k] = z * available.get(k.lower(), 1.0)
        return out

    def to_config(self):
        return {"available_resources": self.available_resources, "board_map": self.board_map}

    @classmethod
    def from_config(cls, config):
        return cls(**config)


class BoardPercentScaler(BaseScaler):
    def __init__(self, available_resources: dict, board_map: dict[str, int]):
        self.available_resources = {}
        self.board_map = board_map
        for board, resources in available_resources.items():
            board = self._get_board_name(board)
            self.available_resources[board] = {}
            for name, value in resources.items():
                self.available_resources[board][f"{name.lower()}"] = float(value)

    def _get_board_name(self, board):
        try:
            board = int(board)
            for name, idx in self.board_map.items():
                if idx == board:
                    return name.lower()
        except Exception:
            if isinstance(board, str):
                return board.lower()
            return board

    def transform(self, target_row, input_row):
        out = {}
        board_name = input_row.get("board")
        board_name = self._get_board_name(board_name)
        if board_name in self.available_resources:
            available = self.available_resources[board_name]
            for k, v in target_row.items():
                if k.lower() in available:
                    out[k] = v / available[k.lower()] * 100.0
        return out

    def inverse(self, target_row, input_row):
        out = {}
        board_name = input_row.get("board")
        board_name = self._get_board_name(board_name)
        if board_name in self.available_resources:
            available = self.available_resources[board_name]
            for k, z in target_row.items():
                if k.lower() in available:
                    out[k] = z * available[k.lower()] / 100.0
        else:
            print("Board not found in available resources:", board_name)
        return out

    def to_config(self):
        return {"available_resources": self.available_resources, "board_map": self.board_map}

    @classmethod
    def from_config(cls, config):
        return cls(**config)


class Log1pScaler(BaseScaler):
    def __init__(self, mu=None, sigma=None):
        self.mu = {} if mu is None else mu
        self.sigma = {} if sigma is None else sigma

    def _fwd(self, y):
        if _is_torch(y):
            return torch.log1p(torch.clamp(y, min=0.0))
        elif isinstance(y, np.ndarray):
            return np.log1p(np.maximum(y, 0.0))
        else:
            return math.log1p(max(0.0, float(y)))

    def _inv(self, z, mu, sig):
        if _is_torch(z):
            mu_t = torch.as_tensor(mu, dtype=z.dtype, device=z.device)
            sig_t = torch.as_tensor(max(sig, 1e-8), dtype=z.dtype, device=z.device)
            return torch.expm1(z * sig_t + mu_t)
        elif isinstance(z, np.ndarray):
            return np.expm1(z * sig + mu)
        else:
            return math.expm1(float(z) * sig + mu)

    def fit(self, dict_list):
        from collections import defaultdict

        by_key = defaultdict(list)
        for d in dict_list:
            for k, v in d.items():
                by_key[k].append(self._fwd(v))
        for k, arr in by_key.items():
            arr = np.asarray(arr, dtype=float)
            arr = arr[~np.isnan(arr)]
            self.mu[k] = 0.0 if arr.size == 0 else float(arr.mean())
            self.sigma[k] = 1.0 if arr.size == 0 else float(arr.std() + 1e-8)

    def transform(self, target_row, input_row):
        out = {}
        for k, v in target_row.items():
            z = self._fwd(v)
            mu = self.mu[k]
            sig = self.sigma[k]
            if _is_torch(z):
                mu_t = torch.as_tensor(mu, dtype=z.dtype, device=z.device)
                sig_t = torch.as_tensor(max(sig, 1e-8), dtype=z.dtype, device=z.device)
                out[k] = (z - mu_t) / sig_t
            elif isinstance(z, np.ndarray):
                out[k] = (z - mu) / sig
            else:
                out[k] = (z - mu) / sig
        return out

    def inverse(self, target_row, input_row):
        out = {}
        for k, z in target_row.items():
            mu = self.mu[k]
            sig = self.sigma[k]
            if z is None:
                out[k] = np.nan
            else:
                out[k] = self._inv(z, mu, sig)
        return out

    def to_config(self):
        return {"mu": self.mu, "sigma": self.sigma}

    @classmethod
    def from_config(cls, config):
        return cls(**config)



import keras
import numpy as np
import tensorflow as tf
from packaging import version

try:
    import torch
except ImportError:
    torch = None

# Check Keras version
if version.parse(keras.__version__).major >= 3:
    from keras import ops as kops
else:
    kops = None


def smape(y_true, y_pred):
    y_true = np.asarray(y_true, dtype=np.float32)
    y_pred = np.asarray(y_pred, dtype=np.float32)

    y_true = y_true.squeeze()
    y_pred = y_pred.squeeze()

    return 200 * np.mean(np.abs(y_true - y_pred) / (np.abs(y_true) + np.abs(y_pred) + 1))


def rmse(y_true, y_pred):
    y_true = np.asarray(y_true, dtype=np.float32)
    y_pred = np.asarray(y_pred, dtype=np.float32)

    y_true = y_true.squeeze()
    y_pred = y_pred.squeeze()

    return np.sqrt(np.mean((y_true - y_pred) ** 2))


class KerasParametricMAPE(tf.keras.metrics.Metric):
    def __init__(self, y_index, name=None, eps=1e-6, **kwargs):
        if name is None:
            name = f"mape_{y_index}"

        super().__init__(name=name, **kwargs)
        self.y_index = y_index
        self.eps = eps
        self.total = self.add_weight(name="total", initializer="zeros")
        self.count = self.add_weight(name="count", initializer="zeros")

    def update_state(self, y_true, y_pred, sample_weight=None):
        y_true_i = y_true[..., self.y_index]
        y_pred_i = y_pred[..., self.y_index]

        if kops is None:
            diff = tf.abs(y_true_i - y_pred_i) / tf.maximum(tf.abs(y_true_i), self.eps)
            mape = 100.0 * diff
            self.total.assign_add(tf.reduce_sum(mape))
            self.count.assign_add(tf.cast(tf.size(mape), tf.float32))
        else:
            diff = kops.absolute(y_true_i - y_pred_i) / kops.maximum(
                kops.absolute(y_true_i), self.eps
            )
            mape = 100.0 * diff
            self.total.assign_add(kops.sum(mape))
            self.count.assign_add(kops.cast(kops.size(mape), tf.float32))

    def result(self):
        return self.total / self.count


class KerasParametricSMAPE(tf.keras.metrics.Metric):
    def __init__(self, y_index, name=None, eps=1e-6, **kwargs):
        if name is None:
            name = f"smape_{y_index}"

        super().__init__(name=name, **kwargs)
        self.y_index = y_index
        self.eps = eps
        self.total = self.add_weight(name="total", initializer="zeros")
        self.count = self.add_weight(name="count", initializer="zeros")

    def update_state(self, y_true, y_pred, sample_weight=None):
        y_true_i = y_true[..., self.y_index]
        y_pred_i = y_pred[..., self.y_index]

        if kops is None:
            numerator = tf.abs(y_true_i - y_pred_i)
            denominator = tf.maximum(tf.abs(y_true_i) + tf.abs(y_pred_i), self.eps)
            smape = 100.0 * numerator / denominator
            self.total.assign_add(tf.reduce_sum(smape))
            self.count.assign_add(tf.cast(tf.size(smape), tf.float32))
        else:
            numerator = kops.absolute(y_true_i - y_pred_i)
            denominator = kops.maximum(kops.absolute(y_true_i) + kops.absolute(y_pred_i), self.eps)
            smape = 100.0 * numerator / denominator
            self.total.assign_add(kops.sum(smape))
            self.count.assign_add(kops.cast(kops.size(smape), tf.float32))

    def result(self):
        return self.total / self.count


class KerasParametricR2(tf.keras.metrics.Metric):
    def __init__(self, y_index, name=None, eps=1e-6, **kwargs):
        if name is None:
            name = f"r2_{y_index}"

        super().__init__(name=name, **kwargs)
        self.y_index = y_index
        self.eps = eps
        self.total = self.add_weight(name="total", initializer="zeros")
        self.count = self.add_weight(name="count", initializer="zeros")

    def update_state(self, y_true, y_pred, sample_weight=None):
        y_true_i = y_true[..., self.y_index]
        y_pred_i = y_pred[..., self.y_index]

        if kops is None:
            mean_true = tf.reduce_mean(y_true_i, axis=-1, keepdims=True)
            ss_res = tf.reduce_sum(tf.square(y_true_i - y_pred_i), axis=-1)
            ss_tot = tf.reduce_sum(tf.square(y_true_i - mean_true), axis=-1)
            ss_tot = tf.maximum(ss_tot, self.eps)
            r2_score = 1.0 - ss_res / ss_tot
            self.total.assign_add(tf.reduce_sum(r2_score))
            self.count.assign_add(tf.cast(tf.size(r2_score), tf.float32))
        else:
            mean_true = kops.mean(y_true_i, axis=-1, keepdims=True)
            ss_res = kops.sum(kops.square(y_true_i - y_pred_i), axis=-1)
            ss_tot = kops.sum(kops.square(y_true_i - mean_true), axis=-1)
            ss_tot = kops.maximum(ss_tot, self.eps)
            r2_score = 1.0 - ss_res / ss_tot
            self.total.assign_add(kops.sum(r2_score))
            self.count.assign_add(kops.cast(kops.size(r2_score), tf.float32))

    def result(self):
        return self.total / self.count


class TorchParametricMAPE(torch.nn.Module):
    def __init__(self, y_index, name="", eps=1e-6, scaler=None, device=None):
        if torch is None:
            raise ImportError(
                "PyTorch is not installed. Please install PyTorch to use this backend."
            )

        super().__init__()
        self.y_index = y_index
        self.eps = eps
        self.name = name if name else f"mape_{y_index}"
        self.scaler = scaler

        if device is None:
            if torch.cuda.is_available():
                device = torch.device("cuda")
            else:
                device = torch.device("cpu")
        self.device = device

        self.total = torch.zeros(1, dtype=torch.float32, device=self.device)
        self.count = torch.zeros(1, dtype=torch.float32, device=self.device)

    def update(self, y_true, y_pred, keys=None):
        y_true_i = y_true[..., self.y_index]
        y_pred_i = y_pred[..., self.y_index]

        if self.scaler and keys:
            key = keys[self.y_index]
            y_true_i = self.scaler.inverse({key: y_true_i})[key]
            y_pred_i = self.scaler.inverse({key: y_pred_i})[key]

        y_true_i = y_true_i.reshape(-1)
        y_pred_i = y_pred_i.reshape(-1)

        diff = torch.abs(y_true_i - y_pred_i) / torch.clamp(torch.abs(y_true_i), min=self.eps)
        mape = 100.0 * torch.mean(diff, dim=-1)

        self.total += torch.sum(mape)
        self.count += torch.tensor(mape.numel(), dtype=torch.float32, device=self.device)

    def result(self):
        if self.count.item() == 0:
            return torch.tensor(0.0, dtype=torch.float32, device=self.device)

        return self.total / self.count

    def reset(self):
        self.total.fill_(0)
        self.count.fill_(0)


class TorchParametricSMAPE(torch.nn.Module):
    def __init__(self, y_index, name="", eps=1e-6, scaler=None, device=None):
        if torch is None:
            raise ImportError(
                "PyTorch is not installed. Please install PyTorch to use this backend."
            )

        super().__init__()
        self.y_index = y_index
        self.eps = eps
        self.name = name if name else f"smape_{y_index}"
        self.scaler = scaler

        if device is None:
            if torch.cuda.is_available():
                device = torch.device("cuda")
            else:
                device = torch.device("cpu")
        self.device = device

        self.total = torch.zeros(1, dtype=torch.float32, device=self.device)
        self.count = torch.zeros(1, dtype=torch.float32, device=self.device)

    def update(self, y_true, y_pred, keys=None):
        y_true_i = y_true[..., self.y_index]
        y_pred_i = y_pred[..., self.y_index]

        if self.scaler and keys:
            key = keys[self.y_index]
            y_true_i = self.scaler.inverse({key: y_true_i})[key]
            y_pred_i = self.scaler.inverse({key: y_pred_i})[key]

        y_true_i = y_true_i.reshape(-1)
        y_pred_i = y_pred_i.reshape(-1)

        numerator = torch.abs(y_true_i - y_pred_i)
        denominator = torch.clamp(torch.abs(y_true_i) + torch.abs(y_pred_i), min=self.eps)
        smape = 100.0 * numerator / denominator

        self.total += torch.sum(smape)
        self.count += torch.tensor(smape.numel(), dtype=torch.float32, device=self.device)
    
    def result(self):
        if self.count.item() == 0:
            return torch.tensor(0.0, dtype=torch.float32, device=self.device)

        return self.total / self.count

    def reset(self):
        self.total.fill_(0)
        self.count.fill_(0)


class TorchParametricR2(torch.nn.Module):
    def __init__(self, y_index, name="", eps=1e-6, scaler=None, device=None):
        if torch is None:
            raise ImportError(
                "PyTorch is not installed. Please install PyTorch to use this backend."
            )

        super().__init__()
        self.y_index = y_index
        self.eps = eps
        self.name = name if name else f"r2_{y_index}"
        self.scaler = scaler

        if device is None:
            if torch.cuda.is_available():
                device = torch.device("cuda")
            else:
                device = torch.device("cpu")
        self.device = device

        self.reset()

    def update(self, y_true, y_pred, keys=None):
        y_true_i = y_true[..., self.y_index]
        y_pred_i = y_pred[..., self.y_index]

        if self.scaler and keys:
            key = keys[self.y_index]
            y_true_i = self.scaler.inverse({key: y_true_i})[key]
            y_pred_i = self.scaler.inverse({key: y_pred_i})[key]

        y_true_i = y_true_i.reshape(-1)
        y_pred_i = y_pred_i.reshape(-1)

        self.y_true_all.append(y_true_i)
        self.y_pred_all.append(y_pred_i)

    def result(self):
        y_true = torch.cat(self.y_true_all)
        y_pred = torch.cat(self.y_pred_all)

        mean_true = torch.mean(y_true)
        ss_res = torch.sum((y_true - y_pred) ** 2)
        ss_tot = torch.sum((y_true - mean_true) ** 2)
        ss_tot = torch.clamp(ss_tot, min=self.eps)

        r2 = 1.0 - ss_res / ss_tot

        return r2

    def reset(self):
        self.y_true_all = []
        self.y_pred_all = []



import keras

try:
    import torch
except ImportError:
    torch = None


def torch_weights_init(m):
    if torch is None:
        raise ImportError(
            "PyTorch is not installed. Please install PyTorch to use this functionality."
        )

    if isinstance(m, torch.nn.Linear):
        torch.nn.init.xavier_uniform_(m.weight)
        if m.bias is not None:
            torch.nn.init.zeros_(m.bias)


def get_optimizer_from_str(optimizer_name, backend, **kwargs):
    """
    Get an optimizer from a string name.

    Args:
        optimizer_name (str): The name of the optimizer.
        backend (str): The backend to use ('keras' or 'torch').
        **kwargs: Additional arguments to pass to the optimizer.

    Returns:
        Optimizer: The optimizer instance.
    """

    optimizer_name = optimizer_name.lower()
    if backend == "keras":
        optimizers = {
            "adam": keras.optimizers.Adam,
            "sgd": keras.optimizers.SGD,
            "rmsprop": keras.optimizers.RMSprop,
            "adagrad": keras.optimizers.Adagrad,
            "adadelta": keras.optimizers.Adadelta,
            "adamax": keras.optimizers.Adamax,
            "nadam": keras.optimizers.Nadam,
        }
    elif backend == "torch":
        if torch is None:
            raise ImportError(
                "PyTorch is not installed. Please install PyTorch to use this backend."
            )

        optimizers = {
            "adam": torch.optim.Adam,
            "sgd": torch.optim.SGD,
            "rmsprop": torch.optim.RMSprop,
            "adagrad": torch.optim.Adagrad,
            "adadelta": torch.optim.Adadelta,
            "adamax": torch.optim.Adamax,
            "nadam": torch.optim.NAdam,
        }
    else:
        raise ValueError(
            f"Unsupported backend \"{backend}\". Supported backends are 'keras' and 'torch'."
        )

    if optimizer_name not in optimizers:
        raise ValueError(
            f"Unsupported optimizer \"{optimizer_name}\". Supported optimizers are {list(optimizers.keys())}."
        )

    return optimizers[optimizer_name](**kwargs)


def get_loss_from_str(loss_name, backend, **kwargs):
    """
    Get a loss function from a string name.

    Args:
        loss_name (str): The name of the loss function.
        backend (str): The backend to use ('keras' or 'torch').

    Returns:
        Loss function: The loss function instance.
    """

    loss_name = loss_name.lower()
    if backend == "keras":
        losses = {
            "mse": keras.losses.MeanSquaredError,
            "msle": keras.losses.MeanSquaredLogarithmicError,
            "mae": keras.losses.MeanAbsoluteError,
            "binary_crossentropy": keras.losses.BinaryCrossentropy,
            "categorical_crossentropy": keras.losses.CategoricalCrossentropy,
            "sparse_categorical_crossentropy": keras.losses.SparseCategoricalCrossentropy,
        }
    elif backend == "torch":
        if torch is None:
            raise ImportError(
                "PyTorch is not installed. Please install PyTorch to use this backend."
            )

        losses = {
            "mse": torch.nn.MSELoss,
            "msle": MSLELoss,
            "mae": torch.nn.L1Loss,
            "binary_crossentropy": torch.nn.BCELoss,
            "categorical_crossentropy": torch.nn.CrossEntropyLoss,
            "sparse_categorical_crossentropy": torch.nn.CrossEntropyLoss,
            "tweedie": TweedieLoss,
        }
    else:
        raise ValueError(
            f"Unsupported backend \"{backend}\". Supported backends are 'keras' and 'torch'."
        )

    if loss_name not in losses:
        raise ValueError(
            f"Unsupported loss \"{loss_name}\". Supported losses are {list(losses.keys())}."
        )

    return losses[loss_name](**kwargs)


class MSLELoss(torch.nn.Module):
    def __init__(self, reduction="mean"):
        super().__init__()
        self.reduction = reduction

    def forward(self, y_true, y_pred):
        loss = (torch.log1p(y_pred) - torch.log1p(y_true)) ** 2

        if self.reduction == "mean":
            return torch.mean(loss)
        elif self.reduction == "sum":
            return torch.sum(loss)
        else:  # no reduction
            return loss


class TweedieLoss(torch.nn.Module):
    def __init__(self, p=1.5, reduction="mean", eps=1e-9):
        super().__init__()
        self.p = p
        self.reduction = reduction
        self.eps = eps

    def forward(self, y_true, y_pred):
        y_true = y_true.clamp_min(0.0)
        y_pred = y_pred.clamp_min(self.eps)
        if self.p == 1.0:  # Poisson limit
            loss = 2.0 * (y_true * torch.log((y_true + self.eps) / y_pred) - (y_true - y_pred))
        elif self.p == 2.0:  # Gamma limit
            loss = 2.0 * (torch.log(y_pred / (y_true + self.eps)) + (y_true - y_pred) / y_pred)
        else:
            loss = 2.0 * (
                (y_true * y_pred.pow(1.0 - self.p)) / (1.0 - self.p)
                - (y_pred.pow(2.0 - self.p)) / (2.0 - self.p)
            )

        if self.reduction == "mean":
            return torch.mean(loss)
        elif self.reduction == "sum":
            return torch.sum(loss)
        else:
            return loss



from collections import OrderedDict

import keras
import numpy as np
import tensorflow as tf

from rule4ml.parsers.utils import camel_keys_to_snake, get_closest_reuse_factor

try:
    import torch
except ImportError:
    torch = None

try:
    import onnx
except ImportError:
    onnx = None


def config_from_keras_model(model, hls_config):
    """
    _summary_

    Args:
        model (_type_): _description_
        hls_config (_type_): _description_

    Returns:
        _type_: _description_
    """

    if hasattr(model, "_build_input_shape") and model._build_input_shape is not None:  # Keras 2
        model_input_shape = model._build_input_shape
    elif hasattr(model, "_build_shapes_dict") and model._build_shapes_dict is not None:  # Keras 3
        model_input_shape = list(model._build_shapes_dict.values())[0]
    else:
        raise AttributeError(
            "Could not get model input shape. Make sure model.build() was called previously."
        )

    dummy_input = tf.random.uniform((1, *model_input_shape[1:]))
    _ = model(dummy_input)

    hls_config = camel_keys_to_snake(hls_config)

    layers_data = []
    for idx, layer in enumerate(model.layers):
        nested_activation = False

        class_name = layer.__class__.__name__
        layer_name = layer.name

        layer_config = layer.get_config()
        layer_weights = layer.get_weights()

        layer_dict = {}
        layer_dict["class_name"] = class_name
        layer_dict["name"] = layer_name

        if class_name.lower() == "activation":
            if idx == 0:
                input_shape = model_input_shape
            else:
                input_shape = layers_data[-1]["output_shape"]

        elif hasattr(
            layer, "input_shape"
        ):  # Keras 2, Sequential and Functional APIs (not including subclassing)
            input_shape = layer.input_shape
        elif hasattr(layer, "_build_input_shape"):  # Keras 2, Subclassed from keras.Model
            input_shape = layer._build_input_shape

        elif hasattr(layer, "batch_shape"):  # Keras 3, InputLayer
            input_shape = layer.batch_shape
        elif (  # Keras 3, other layers (best way we found, not documented)
            hasattr(layer, "_build_shapes_dict") and layer._build_shapes_dict is not None
        ):
            input_shape = list(layer._build_shapes_dict.values())[0]

        else:
            raise AttributeError(
                f"Could not get the input shape for layer {layer.name}. Make sure model.build() was called previously."
            )

        input_shape = np.asarray(input_shape).flatten()
        input_shape = tuple(input_shape)
        layer_dict["input_shape"] = input_shape

        if class_name == "InputLayer":  # Same input and output shape for InputLayer
            output_shape = input_shape
        elif hasattr(layer, "output_shape"):  # Keras 2
            output_shape = layer.output_shape
        else:  # Keras 3, layers other than InputLayer
            output_shape = layer.compute_output_shape(input_shape)

        output_shape = tuple(output_shape)
        layer_dict["output_shape"] = tuple(output_shape)

        # Tracking inbound layers can be useful for add/concatenate layers
        if hasattr(layer, "inbound_nodes"):
            inbound_nodes = layer.inbound_nodes
            inbound_layers = []
            for node in inbound_nodes:
                if not isinstance(node.inbound_layers, (list, tuple)):
                    inbound_layers.append(node.inbound_layers)
                else:
                    inbound_layers += node.inbound_layers
            layer_dict["inbound_layers"] = [layer.name for layer in inbound_layers]

        parameter_count = 0
        for weight_group in layer_weights:
            parameter_count += np.size(weight_group)

        parameter_count = int(parameter_count)
        layer_dict["parameters"] = parameter_count

        trainable_parameter_count = 0
        for var_group in layer.trainable_variables:
            trainable_parameter_count += np.size(var_group)

        trainable_parameter_count = int(trainable_parameter_count)
        layer_dict["trainable_parameters"] = trainable_parameter_count

        if class_name in ["Dense", "QDense"]:
            layer_dict["neurons"] = int(layer_config["units"])
            layer_dict["use_bias"] = layer_config["use_bias"]

        elif class_name in ["Conv1D", "Conv2D", "QConv1D", "QConv2D"]:
            layer_dict["channels"] = int(input_shape[-1])
            layer_dict["filters"] = int(layer_config["filters"])
            layer_dict["kernel_size"] = tuple([int(x) for x in layer_config["kernel_size"]])
            layer_dict["strides"] = tuple([int(x) for x in layer_config["strides"]])
            layer_dict["padding"] = layer_config["padding"]
            layer_dict["use_bias"] = layer_config["use_bias"]

        elif class_name == "Dropout":
            layer_dict["dropout_rate"] = layer_config["rate"]

        if "activation" in layer_config and layer_config["activation"] != "linear":
            if class_name in ["Activation", "QActivation"]:
                layer_dict["activation"] = layer_config["activation"]
            else:
                nested_activation = True

        dtype = layer_config["dtype"]
        if isinstance(dtype, dict):
            dtype = dtype["config"]["name"]
        layer_dict["dtype"] = dtype

        reuse_factor = hls_config["model"]["reuse_factor"]
        if "layer_type" in hls_config and class_name in hls_config["layer_type"]:
            reuse_factor = hls_config["layer_type"][class_name].get("reuse_factor", reuse_factor)
        if "layer_name" in hls_config and layer.name in hls_config["layer_name"]:
            reuse_factor = hls_config["layer_name"][layer.name].get("reuse_factor", reuse_factor)

        layer_dict["reuse_factor"] = reuse_factor
        if class_name in ["Dense", "QDense"]:
            n_in = np.prod([x for x in input_shape if x is not None])
            n_out = np.prod([x for x in output_shape if x is not None])
            layer_dict["reuse_factor"] = get_closest_reuse_factor(n_in, n_out, reuse_factor)
        elif class_name in ["Conv1D", "Conv2D", "QConv1D", "QConv2D"]:
            n_in = layer_dict["channels"] * np.prod(layer_dict["kernel_size"])
            n_out = layer_dict["filters"]
            layer_dict["reuse_factor"] = get_closest_reuse_factor(n_in, n_out, reuse_factor)

        layers_data.append(layer_dict)

        if (
            nested_activation
        ):  # activation function wrapped in a layer other than "(Q)Activation", example: Dense(units=32, activation="relu")
            activation_dict = {}
            activation_dict["class_name"] = "Activation"

            activation_dict["input_shape"] = output_shape
            activation_dict["output_shape"] = output_shape

            activation_dict["activation"] = layer_config["activation"]

            activation_dict["parameters"] = 0
            activation_dict["trainable_parameters"] = 0

            activation_dict["dtype"] = layer_config["dtype"]
            activation_dict["reuse_factor"] = reuse_factor

            layers_data.append(activation_dict)

    return layers_data


def keras_model_from_config(model_config):
    layer_dict = {}

    x = inputs = None
    for layer_config in model_config:
        class_name = layer_config["class_name"]
        input_shape = layer_config.get("input_shape", [])[1:]

        if class_name == "InputLayer":
            x = inputs = keras.layers.Input(shape=input_shape)

        elif class_name == "Dense":
            dense_layer = keras.layers.Dense(
                units=layer_config["neurons"],
                use_bias=layer_config["use_bias"],
                dtype=layer_config["dtype"],
            )
            layer_dict[dense_layer.name] = dense_layer
            x = dense_layer(x)

        elif class_name == "Activation":
            activation = layer_config["activation"]
            activation_layer = keras.layers.Activation(activation)
            layer_dict[activation_layer.name] = activation_layer
            x = activation_layer(x)

        elif class_name == "Dropout":
            dropout_rate = layer_config.get(
                "dropout_rate", np.random.default_rng().uniform(0.1, 0.8)
            )
            dropout_layer = keras.layers.Dropout(dropout_rate)
            layer_dict[dropout_layer.name] = dropout_layer
            x = dropout_layer(x)

        elif class_name in ["Add", "Concatenate"]:
            inbound_names = layer_config["inbound_layers"]
            inbound_layers = [layer_dict[name] for name in inbound_names]

            skip_layer = getattr(keras.layers, class_name)()
            layer_dict[skip_layer.name] = skip_layer
            x = skip_layer(inbound_layers)

        else:
            keras_layer = getattr(keras.layers, class_name)()
            layer_dict[keras_layer.name] = keras_layer
            x = keras_layer(x)

    model = keras.Model(inputs=inputs, outputs=x)
    return model


def config_from_torch_model(model, hls_config):
    """
    _summary_

    Args:
        model (_type_): _description_
        hls_config (_type_): _description_

    Returns:
        _type_: _description_
    """

    if torch is None:
        raise ImportError("Failed to import module torch.")

    torch_layer_mapping = {
        "Linear": "Dense",
        "Conv2d": "Conv2D",
        "add": "Add",
        "cat": "Concatenate",
        "BatchNorm1d": "BatchNormalization",
        "BatchNorm2d": "BatchNormalization",
    }

    hls_config = camel_keys_to_snake(hls_config)

    layers_data = OrderedDict()
    model_layers = get_torch_layers(model)
    traced_model = torch.fx.symbolic_trace(model)
    module_count = 0
    for idx, node in enumerate(traced_model.graph.nodes):
        if node.op == "output":
            continue

        layer_dict = {}

        if node.op == "placeholder":
            class_name = "InputLayer"

            reuse_factor = hls_config["model"]["reuse_factor"]
            if "layer_type" in hls_config and class_name in hls_config["layer_type"]:
                reuse_factor = hls_config["layer_type"][class_name].get(
                    "reuse_factor", reuse_factor
                )
            if "layer_name" in hls_config and node.name in hls_config["layer_name"]:
                reuse_factor = hls_config["layer_name"][node.name].get("reuse_factor", reuse_factor)

            layer_dict = {
                "class_name": class_name,
                "input_shape": (None,),
                "output_shape": (None,),
                "parameters": 0,
                "trainable_parameters": 0,
                "dtype": "float32",
                "reuse_factor": reuse_factor,
            }

        elif node.op == "call_module":
            layer = model_layers[module_count]

            class_name = layer.__class__.__name__
            mapped_class_name = torch_layer_mapping.get(class_name, class_name)
            layer_dict["class_name"] = mapped_class_name

            if mapped_class_name in ["Dense"]:
                layer_dict["input_shape"] = (None, layer.in_features)
                layer_dict["output_shape"] = (None, layer.out_features)

                use_bias = layer.bias is not None
                layer_dict["parameters"] = (layer.in_features + int(use_bias)) * layer.out_features
                layer_dict["trainable_parameters"] = layer_dict["parameters"]

                layer_dict["neurons"] = layer_dict["output_shape"][-1]
                layer_dict["use_bias"] = use_bias

                layer_dict["dtype"] = ".".join(str(layer.weight.dtype).split(".")[1:])

            elif module_count > 0:
                layers_data_values = list(layers_data.values())

                layer_dict["input_shape"] = layers_data_values[idx - 1]["output_shape"]
                layer_dict["output_shape"] = layer_dict["input_shape"]

                if mapped_class_name in ["BatchNormalization"]:
                    input_size = np.prod([x for x in layer_dict["input_shape"] if x is not None])
                    layer_dict["parameters"] = 4 * input_size
                    layer_dict["trainable_parameters"] = 2 * input_size
                else:
                    layer_dict["parameters"] = 0
                    layer_dict["trainable_parameters"] = 0

                layer_dict["dtype"] = layers_data_values[idx - 1]["dtype"]

            else:
                raise TypeError(f"Model has unexpected first layer of type {class_name}.")

            reuse_factor = hls_config["model"]["reuse_factor"]
            if "layer_type" in hls_config and mapped_class_name in hls_config["layer_type"]:
                reuse_factor = hls_config["layer_type"][mapped_class_name].get(
                    "reuse_factor", reuse_factor
                )
            if "layer_name" in hls_config and node.name in hls_config["layer_name"]:
                reuse_factor = hls_config["layer_name"][node.name].get("reuse_factor", reuse_factor)

            layer_dict["reuse_factor"] = reuse_factor
            if mapped_class_name in ["Dense"]:
                n_in = np.prod([x for x in layer_dict["input_shape"] if x is not None])
                n_out = np.prod([x for x in layer_dict["output_shape"] if x is not None])

                layer_dict["reuse_factor"] = get_closest_reuse_factor(n_in, n_out, reuse_factor)

            module_count += 1

        elif node.op == "call_function":
            class_name = node.name.split("_")
            if class_name[-1].isdigit():
                class_name = class_name[:-1]
            class_name = "_".join(class_name)
            layer_dict["class_name"] = class_name

            layers_data_values = list(layers_data.values())

            input_shapes = [
                layers_data[input_node.name]["output_shape"] for input_node in node.all_input_nodes
            ]
            layer_dict["input_shape"] = input_shapes

            if node.name == "cat":
                dim = node.kwargs["dim"]
                output_shape = list(input_shapes[0])
                output_shape[dim] = sum([x[dim] for x in input_shapes])

                layer_dict["output_shape"] = tuple(output_shape)
            else:
                layer_dict["output_shape"] = input_shapes[0]

            layer_dict["parameters"] = 0
            layer_dict["trainable_parameters"] = 0

            layer_dict["dtype"] = layers_data_values[idx - 1]["dtype"]
            layer_dict["reuse_factor"] = reuse_factor

            layer_dict["class_name"] = torch_layer_mapping.get(class_name, class_name)

        layers_data[node.name] = layer_dict

    layers_data = list(layers_data.values())

    model_input_shape = layers_data[1]["input_shape"]
    layers_data[0]["input_shape"] = model_input_shape
    layers_data[0]["output_shape"] = model_input_shape
    layers_data[0]["dtype"] = layers_data[1]["dtype"]

    return layers_data


def config_from_onnx_model(model, hls_config):
    """_summary_

    Args:
        model (_type_): _description_
        hls_config (_type_): _description_

    Raises:
        ImportError: _description_
    """

    if onnx is None:
        raise ImportError("Failed to import module onnx.")

    layers_data = []
    return layers_data


def get_torch_layers(model):
    layers = []

    model_children = list(model.children())
    if model_children == []:
        return [model]

    for child in model_children:
        layers += get_torch_layers(child)

    return layers



import json
import math
import os
import re
from bisect import bisect_left


def camel_keys_to_snake(d: dict, recursive=True):
    """
    Transforms dictionary keys from CamelCase to snake_case format.

    Args:
        d (dict): Targeted dictionary for keys conversion.
        recursive (bool, optional): Whether to apply to nested dictionaries. Defaults to True.

    Returns:
        dict: Transformed dictionary.
    """

    if not isinstance(d, dict):
        return d

    new_dict = {}
    for key, value in d.items():
        new_key = re.sub(r"(?<!^)(?=[A-Z])", "_", key).lower()
        if recursive and isinstance(value, dict):
            new_dict[new_key] = camel_keys_to_snake(value, recursive=True)
        else:
            new_dict[new_key] = value

    return new_dict


def to_lower_keys(d: dict, recursive=True):
    """
    Transforms dictionary keys to lower case.

    Args:
        d (dict): Targeted dictionary for keys conversion.
        recursive (bool, optional): Whether to apply to nested dictionaries. Defaults to True.

    Returns:
        dict: Transformed dictionary.
    """

    if not isinstance(d, dict):
        return d

    new_dict = {}
    for key, value in d.items():
        new_key = key.lower()
        if recursive and isinstance(value, dict):
            new_dict[new_key] = to_lower_keys(value, recursive=True)
        else:
            new_dict[new_key] = value

    return new_dict


def unwrap_nested_dicts(d, prefix=None):
    """
    _summary_

    Args:
        d (_type_): _description_
        prefix (_type_, optional): _description_. Defaults to None.

    Returns:
        _type_: _description_
    """

    if not isinstance(d, dict):
        return d

    items = []
    for key, value in d.items():
        new_key = f"{prefix}_{key}" if prefix is not None else key
        if isinstance(value, dict):
            items += unwrap_nested_dicts(value, prefix=new_key).items()
        else:
            items.append((new_key, value))

    return dict(items)


def get_activation_name(activation):
    match = re.match(r"quantized_(\w+)(?:\(?[^)]*\)?)", activation)
    if match:
        return match.group(1)
    return activation


def fixed_precision_to_bit_width(precision: str):
    """
    _summary_

    Args:
        precision (str): _description_

    Raises:
        ValueError: _description_

    Returns:
        _type_: _description_
    """

    match = re.match(r"(ap_)?fixed<(\d+),\s*(\d+)>", precision.lower())
    if not match:
        raise ValueError(f"Invalid format: {precision}, expecting \"ap_fixed<a, b>\"")

    groups = match.groups()
    if len(groups) == 3:
        groups = groups[1:]

    total_bits, integer_bits = map(str, groups)
    return int(total_bits), int(integer_bits)


def get_supported_boards():
    """
    _summary_

    Returns:
        _type_: _description_
    """

    with open(os.path.join(os.path.dirname(__file__), "supported_boards.json")) as json_file:
        supported_boards = json.load(json_file)

    return supported_boards


def get_board_from_part(part_number):
    """
    _summary_

    Args:
        part_number (_type_): _description_

    Raises:
        ValueError: _description_

    Returns:
        _type_: _description_
    """

    supported_boards = get_supported_boards()

    for key, value in supported_boards.items():
        if value["part"] == part_number:
            return key

    raise ValueError(f"Part number {part_number} not supported.")


def get_part_from_board(board_name):
    """
    _summary_

    Args:
        board_name (_type_): _description_

    Raises:
        ValueError: _description_

    Returns:
        _type_: _description_
    """

    supported_boards = get_supported_boards()

    if board_name not in supported_boards:
        raise ValueError(f"Board {board_name} not supported.")

    return supported_boards[board_name]["part"]


# https://github.com/fastmachinelearning/hls4ml/blob/main/hls4ml/backends/fpga/fpga_backend.py#L262
def _validate_reuse_factor(n_in, n_out, rf):
    multfactor = min(n_in, rf)
    multiplier_limit = int(math.ceil((n_in * n_out) / float(multfactor)))
    _assert = ((multiplier_limit % n_out) == 0) or (rf >= n_in)
    _assert = _assert and (((rf % n_in) == 0) or (rf < n_in))
    _assert = _assert and (((n_in * n_out) % rf) == 0)

    return _assert


# https://github.com/fastmachinelearning/hls4ml/blob/main/hls4ml/backends/fpga/fpga_backend.py#L277
def get_closest_reuse_factor(n_in, n_out, chosen_rf):
    """
    Returns closest value to chosen_rf.
    If two numbers are equally close, return the smallest number.
    """

    max_rf = n_in * n_out
    valid_reuse_factors = []
    for rf in range(1, max_rf + 1):
        _assert = _validate_reuse_factor(n_in, n_out, rf)
        if _assert:
            valid_reuse_factors.append(rf)
    valid_rf = sorted(valid_reuse_factors)

    pos = bisect_left(valid_rf, chosen_rf)
    if pos == 0:
        return valid_rf[0]
    if pos == len(valid_rf):
        return valid_rf[-1]
    before = valid_rf[pos - 1]
    after = valid_rf[pos]
    if after - chosen_rf < chosen_rf - before:
        return after
    else:
        return before
